from datetime import date
from datetime import datetime
from datetime import timezone
import json
from uuid import UUID

import pytest

import falcon
from falcon.errors import HTTPInvalidParam
import falcon.testing as testing
from falcon.util import deprecation


class Resource(testing.SimpleTestResource):
    @falcon.before(testing.capture_responder_args)
    @falcon.before(testing.set_resp_defaults)
    def on_put(self, req, resp, **kwargs):
        pass

    @falcon.before(testing.capture_responder_args)
    @falcon.before(testing.set_resp_defaults)
    def on_patch(self, req, resp, **kwargs):
        pass

    @falcon.before(testing.capture_responder_args)
    @falcon.before(testing.set_resp_defaults)
    def on_delete(self, req, resp, **kwargs):
        pass

    @falcon.before(testing.capture_responder_args)
    @falcon.before(testing.set_resp_defaults)
    def on_head(self, req, resp, **kwargs):
        pass

    @falcon.before(testing.capture_responder_args)
    @falcon.before(testing.set_resp_defaults)
    def on_options(self, req, resp, **kwargs):
        pass


@pytest.fixture
def resource():
    return Resource()


@pytest.fixture
def client(asgi, util):
    app = util.create_app(asgi)
    if not asgi:
        with pytest.warns(deprecation.DeprecatedWarning):
            app.req_options.auto_parse_form_urlencoded = True

    return testing.TestClient(app)


def simulate_request_get_query_params(client, path, query_string, **kwargs):
    return client.simulate_request(path=path, query_string=query_string, **kwargs)


def simulate_request_post_query_params(client, path, query_string, **kwargs):
    if client.app._ASGI:
        pytest.skip(
            'The ASGI implementation does not support '
            'RequestOptions.auto_parse_form_urlencoded'
        )

    headers = kwargs.setdefault('headers', {})
    headers['Content-Type'] = 'application/x-www-form-urlencoded'
    if 'method' not in kwargs:
        kwargs['method'] = 'POST'
    return client.simulate_request(path=path, body=query_string, **kwargs)


@pytest.fixture(
    scope='session',
    params=[
        simulate_request_get_query_params,
        simulate_request_post_query_params,
    ],
)
def simulate_request(request):
    return request.param


class TestQueryParams:
    def test_none(self, simulate_request, client, resource):
        query_string = ''
        client.app.add_route('/', resource)  # TODO: DRY up this setup logic
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        store = {}
        assert req.get_param('marker') is None
        assert req.get_param('limit', store) is None
        assert 'limit' not in store
        assert req.get_param_as_int('limit') is None
        assert req.get_param_as_float('limit') is None
        assert req.get_param_as_bool('limit') is None
        assert req.get_param_as_list('limit') is None

    def test_default(self, simulate_request, client, resource):
        default = 'foobar'
        query_string = ''
        client.app.add_route('/', resource)  # TODO: DRY up this setup logic
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        store = {}
        assert req.get_param('marker', default=default) == 'foobar'
        assert req.get_param('limit', store, default=default) == 'foobar'
        assert 'limit' not in store
        assert req.get_param_as_int('limit', default=default) == 'foobar'
        assert req.get_param_as_float('limit', default=default) == 'foobar'
        assert req.get_param_as_bool('limit', default=default) == 'foobar'
        assert req.get_param_as_list('limit', default=default) == 'foobar'

    def test_blank(self, simulate_request, client, resource):
        query_string = 'marker='
        client.app.add_route('/', resource)
        client.app.req_options.keep_blank_qs_values = False
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        assert req.get_param('marker') is None

        store = {}
        assert req.get_param('marker', store=store) is None
        assert 'marker' not in store

    def test_simple(self, simulate_request, client, resource):
        query_string = 'marker=deadbeef&limit=25'
        client.app.add_route('/', resource)
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        store = {}
        assert req.get_param('marker', store=store) or 'nada' == 'deadbeef'
        assert req.get_param('limit', store=store) or '0' == '25'

        assert store['marker'] == 'deadbeef'
        assert store['limit'] == '25'

    def test_percent_encoded(self, simulate_request, client, resource):
        query_string = 'id=23,42&q=%e8%b1%86+%e7%93%a3'
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = True
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        # NOTE(kgriffs): For lists, get_param will return one of the
        # elements, but which one it will choose is undefined.
        assert req.get_param('id') in ['23', '42']

        assert req.get_param_as_list('id', int) == [23, 42]
        assert req.get_param('q') == '\u8c46 \u74e3'

    def test_option_auto_parse_qs_csv_simple_false(
        self, simulate_request, client, resource
    ):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = False

        query_string = 'id=23,42,,&id=2'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.params['id'] == ['23,42,,', '2']
        assert req.get_param('id') in ['23,42,,', '2']
        assert req.get_param_as_list('id') == ['23,42,,', '2']

    def test_option_auto_parse_qs_csv_simple_true(
        self, simulate_request, client, resource
    ):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = True
        client.app.req_options.keep_blank_qs_values = False

        query_string = 'id=23,42,,&id=2'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.params['id'] == ['23', '42', '2']
        assert req.get_param('id') in ['23', '42', '2']
        assert req.get_param_as_list('id', int) == [23, 42, 2]

    def test_option_auto_parse_qs_csv_multiple_fields_false(
        self, simulate_request, client, resource
    ):
        client.app.add_route('/', resource)

        query_string = 't=1,2&t=3,4'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.params['t'] == ['1,2', '3,4']
        assert req.get_param('t') in ['1,2', '3,4']
        assert req.get_param_as_list('t') == ['1,2', '3,4']

    @pytest.mark.parametrize(
        'qs, keep_blank, expected',
        [
            ('t=1&t=3,4', False, ['1', '3', '4']),
            ('t=1&t=2&t=3,4', False, ['1', '2', '3', '4']),
            ('t=1,2&t=3,4', False, ['1', '2', '3', '4']),
            ('t=1,,2&t=3,4', False, ['1', '2', '3', '4']),
            ('t=1,,2&t=3,4', True, ['1', '', '2', '3', '4']),
            ('t=1,2&t=3,4,,5', False, ['1', '2', '3', '4', '5']),
            ('t=1&t=,1,4,,5', False, ['1', '1', '4', '5']),
            ('t=1&t=,1,4,,5', True, ['1', '', '1', '4', '', '5']),
            (
                't=1&t=,1,4,,5&t=a,b,c',
                True,
                ['1', '', '1', '4', '', '5', 'a', 'b', 'c'],
            ),
        ],
    )
    def test_option_auto_parse_qs_csv_multiple_fields_true(
        self,
        simulate_request,
        client,
        resource,
        qs,
        keep_blank,
        expected,
    ):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = True
        client.app.req_options.keep_blank_qs_values = keep_blank

        query_string = qs
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.params['t'] == expected
        assert req.get_param('t') in expected
        assert req.get_param_as_list('t') == expected

    def test_option_auto_parse_qs_csv_complex_false(
        self, simulate_request, client, resource
    ):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = False
        client.app.req_options.keep_blank_qs_values = False

        encoded_json = '%7B%22msg%22:%22Testing%201,2,3...%22,%22code%22:857%7D'
        decoded_json = '{"msg":"Testing 1,2,3...","code":857}'

        query_string = (
            'colors=red,green,blue&limit=1'
            '&list-ish1=f,,x&list-ish2=,0&list-ish3=a,,,b'
            '&empty1=&empty2=,&empty3=,,'
            '&thing=' + encoded_json
        )

        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.get_param('colors') in 'red,green,blue'
        assert req.get_param_as_list('colors') == ['red,green,blue']

        assert req.get_param_as_list('limit') == ['1']

        assert req.get_param_as_list('empty1') is None
        assert req.get_param_as_list('empty2') == [',']
        assert req.get_param_as_list('empty3') == [',,']

        assert req.get_param_as_list('list-ish1') == ['f,,x']
        assert req.get_param_as_list('list-ish2') == [',0']
        assert req.get_param_as_list('list-ish3') == ['a,,,b']

        assert req.get_param('thing') == decoded_json

    def test_default_auto_parse_csv_behaviour(self, simulate_request, client, resource):
        client.app.add_route('/', resource=resource)
        query_string = 'id=1,2,,&id=3'

        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        assert req.get_param('id') == '3'
        assert req.get_param_as_list('id') == ['1,2,,', '3']

    def test_bad_percentage(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'x=%%20%+%&y=peregrine&z=%a%z%zz%1%20e'
        response = simulate_request(client=client, path='/', query_string=query_string)
        assert response.status == falcon.HTTP_200

        req = resource.captured_req
        assert req.get_param('x') == '% % %'
        assert req.get_param('y') == 'peregrine'
        assert req.get_param('z') == '%a%z%zz%1 e'

    def test_allowed_names(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        client.app.req_options.keep_blank_qs_values = False
        query_string = (
            'p=0&p1=23&2p=foo&some-thing=that&blank=&'
            'some_thing=x&-bogus=foo&more.things=blah&'
            '_thing=42&_charset_=utf-8'
        )
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        assert req.get_param('p') == '0'
        assert req.get_param('p1') == '23'
        assert req.get_param('2p') == 'foo'
        assert req.get_param('some-thing') == 'that'
        assert req.get_param('blank') is None
        assert req.get_param('some_thing') == 'x'
        assert req.get_param('-bogus') == 'foo'
        assert req.get_param('more.things') == 'blah'
        assert req.get_param('_thing') == '42'
        assert req.get_param('_charset_') == 'utf-8'

    @pytest.mark.parametrize(
        'method_name',
        [
            'get_param',
            'get_param_as_int',
            'get_param_as_float',
            'get_param_as_uuid',
            'get_param_as_bool',
            'get_param_as_list',
        ],
    )
    def test_required(self, simulate_request, client, resource, method_name):
        client.app.add_route('/', resource)
        query_string = ''
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        try:
            getattr(req, method_name)('marker', required=True)
            pytest.fail('falcon.HTTPMissingParam not raised')
        except falcon.HTTPMissingParam as ex:
            assert isinstance(ex, falcon.HTTPBadRequest)
            assert ex.title == 'Missing parameter'
            expected_desc = 'The "marker" parameter is required.'
            assert ex.description == expected_desc

    def test_int(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&limit=25'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        try:
            req.get_param_as_int('marker')
        except Exception as ex:
            assert isinstance(ex, falcon.HTTPBadRequest)
            assert isinstance(ex, falcon.HTTPInvalidParam)
            assert ex.title == 'Invalid parameter'
            expected_desc = (
                'The "marker" parameter is invalid. The value must be an integer.'
            )
            assert ex.description == expected_desc

        assert req.get_param_as_int('limit') == 25

        store = {}
        assert req.get_param_as_int('limit', store=store) == 25
        assert store['limit'] == 25

        assert req.get_param_as_int('limit', min_value=1, max_value=50) == 25

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('limit', min_value=0, max_value=10)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('limit', min_value=0, max_value=24)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('limit', min_value=30, max_value=24)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('limit', min_value=30, max_value=50)

        assert req.get_param_as_int('limit', min_value=1) == 25

        assert req.get_param_as_int('limit', max_value=50) == 25

        assert req.get_param_as_int('limit', max_value=25) == 25

        assert req.get_param_as_int('limit', max_value=26) == 25

        assert req.get_param_as_int('limit', min_value=25) == 25

        assert req.get_param_as_int('limit', min_value=24) == 25

        assert req.get_param_as_int('limit', min_value=-24) == 25

    def test_int_neg(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&pos=-7'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        assert req.get_param_as_int('pos') == -7

        assert req.get_param_as_int('pos', min_value=-10, max_value=10) == -7

        assert req.get_param_as_int('pos', max_value=10) == -7

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('pos', min_value=-6, max_value=0)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('pos', min_value=-6)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('pos', min_value=0, max_value=10)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_int('pos', min_value=0, max_value=10)

    def test_float(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&limit=25.1'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        try:
            req.get_param_as_float('marker')
        except Exception as ex:
            assert isinstance(ex, falcon.HTTPBadRequest)
            assert isinstance(ex, falcon.HTTPInvalidParam)
            assert ex.title == 'Invalid parameter'
            expected_desc = (
                'The "marker" parameter is invalid. The value must be a float.'
            )
            assert ex.description == expected_desc

        assert req.get_param_as_float('limit') == 25.1

        store = {}
        assert req.get_param_as_float('limit', store=store) == 25.1
        assert store['limit'] == 25.1

        assert req.get_param_as_float('limit', min_value=1, max_value=50) == 25.1

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('limit', min_value=0, max_value=10)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('limit', min_value=0, max_value=24)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('limit', min_value=30, max_value=24)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('limit', min_value=30, max_value=50)

        assert req.get_param_as_float('limit', min_value=1) == 25.1

        assert req.get_param_as_float('limit', max_value=50) == 25.1

        assert req.get_param_as_float('limit', max_value=25.1) == 25.1

        assert req.get_param_as_float('limit', max_value=26) == 25.1

        assert req.get_param_as_float('limit', min_value=25) == 25.1

        assert req.get_param_as_float('limit', min_value=24) == 25.1

        assert req.get_param_as_float('limit', min_value=-24) == 25.1

    def test_float_neg(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&pos=-7.1'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        assert req.get_param_as_float('pos') == -7.1

        assert req.get_param_as_float('pos', min_value=-10, max_value=10) == -7.1

        assert req.get_param_as_float('pos', max_value=10) == -7.1

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('pos', min_value=-6, max_value=0)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('pos', min_value=-6)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('pos', min_value=0, max_value=10)

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_float('pos', min_value=0, max_value=10)

    def test_uuid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = (
            'marker1=8d76b7b3-d0dd-46ca-ad6e-3989dcd66959&'
            'marker2=64be949b-3433-4d36-a4a8-9f19d352fee8&'
            'marker2=8D76B7B3-d0dd-46ca-ad6e-3989DCD66959&'
            'short=4be949b-3433-4d36-a4a8-9f19d352fee8'
        )
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        expected_uuid = UUID('8d76b7b3-d0dd-46ca-ad6e-3989dcd66959')
        assert req.get_param_as_uuid('marker1') == expected_uuid
        assert req.get_param_as_uuid('marker2') == expected_uuid
        assert req.get_param_as_uuid('marker3') is None
        assert req.get_param_as_uuid('marker3', required=False) is None

        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_uuid('short')

        store = {}
        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_uuid('marker3', required=True, store=store)

        assert not store
        assert req.get_param_as_uuid('marker1', store=store)
        assert store['marker1'] == expected_uuid

    def test_boolean(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        client.app.req_options.keep_blank_qs_values = False
        query_string = '&'.join(
            [
                'echo=true',
                'doit=false',
                'bogus=bar',
                'bogus2=foo',
                't1=True',
                'f1=False',
                't2=yes',
                'f2=no',
                't3=y',
                'f3=n',
                't4=t',
                'f4=f',
                'blank',
                'one=1',
                'zero=0',
                'checkbox1=on',
                'checkbox2=off',
            ]
        )
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        with pytest.raises(falcon.HTTPBadRequest):
            req.get_param_as_bool('bogus')

        try:
            req.get_param_as_bool('bogus2')
        except Exception as ex:
            assert isinstance(ex, falcon.HTTPInvalidParam)
            assert ex.title == 'Invalid parameter'
            expected_desc = (
                'The "bogus2" parameter is invalid. '
                'The value of the parameter must be "true" '
                'or "false".'
            )
            assert ex.description == expected_desc

        assert req.get_param_as_bool('echo') is True
        assert req.get_param_as_bool('doit') is False

        for i in range(1, 5):
            assert req.get_param_as_bool('t' + str(i)) is True
            assert req.get_param_as_bool('f' + str(i)) is False

        assert req.get_param_as_bool('one') is True
        assert req.get_param_as_bool('zero') is False
        assert req.get_param('blank') is None

        assert req.get_param_as_bool('checkbox1') is True
        assert req.get_param_as_bool('checkbox2') is False

        store = {}
        assert req.get_param_as_bool('echo', store=store) is True
        assert store['echo'] is True

    def test_boolean_blank(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        simulate_request(client=client, path='/', query_string='blank&blank2=')

        req = resource.captured_req
        assert req.get_param('blank') == ''
        assert req.get_param('blank2') == ''

        for param_name in ('blank', 'blank2'):
            assert req.get_param_as_bool(param_name) is True
            assert req.get_param_as_bool(param_name, blank_as_true=True) is True
            assert req.get_param_as_bool(param_name, blank_as_true=False) is False

        assert req.get_param_as_bool('nichts') is None
        assert req.get_param_as_bool('nichts', default=None) is None
        assert req.get_param_as_bool('nichts', default=False) is False
        assert req.get_param_as_bool('nichts', default=True) is True

    def test_list_type(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = True
        client.app.req_options.keep_blank_qs_values = False
        query_string = (
            'colors=red,green,blue&limit=1'
            '&list-ish1=f,,x&list-ish2=,0&list-ish3=a,,,b'
            '&empty1=&empty2=,&empty3=,,'
            '&thing_one=1,,3'
            '&thing_two=1&thing_two=&thing_two=3'
        )
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        # NOTE(kgriffs): For lists, get_param will return one of the
        # elements, but which one it will choose is undefined.
        assert req.get_param('colors') in ('red', 'green', 'blue')

        assert req.get_param_as_list('colors') == ['red', 'green', 'blue']
        assert req.get_param_as_list('limit') == ['1']
        assert req.get_param_as_list('marker') is None

        assert req.get_param_as_list('empty1') is None
        assert req.get_param_as_list('empty2') == []
        assert req.get_param_as_list('empty3') == []

        assert req.get_param_as_list('list-ish1') == ['f', 'x']

        # Ensure that '0' doesn't get translated to None
        assert req.get_param_as_list('list-ish2') == ['0']

        # Ensure that '0' doesn't get translated to None
        assert req.get_param_as_list('list-ish3') == ['a', 'b']

        # Ensure consistency between list conventions
        assert req.get_param_as_list('thing_one') == ['1', '3']
        assert req.get_param_as_list('thing_one') == req.get_param_as_list('thing_two')

        store = {}
        assert req.get_param_as_list('limit', store=store) == ['1']
        assert store['limit'] == ['1']

    def test_list_type_blank(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = (
            'colors=red,green,blue&limit=1'
            '&list-ish1=f,,x&list-ish2=,0&list-ish3=a,,,b'
            '&empty1=&empty2=,&empty3=,,'
            '&thing_one=1,,3'
            '&thing_two=1&thing_two=&thing_two=3'
            '&empty4=&empty4&empty4='
            '&empty5&empty5&empty5'
        )
        client.app.req_options.keep_blank_qs_values = True
        client.app.req_options.auto_parse_qs_csv = True
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        # NOTE(kgriffs): For lists, get_param will return one of the
        # elements, but which one it will choose is undefined.
        assert req.get_param('colors') in ('red', 'green', 'blue')

        assert req.get_param_as_list('colors') == ['red', 'green', 'blue']
        assert req.get_param_as_list('limit') == ['1']
        assert req.get_param_as_list('marker') is None

        assert req.get_param_as_list('empty1') == ['']
        assert req.get_param_as_list('empty2') == ['', '']
        assert req.get_param_as_list('empty3') == ['', '', '']

        assert req.get_param_as_list('list-ish1') == ['f', '', 'x']

        # Ensure that '0' doesn't get translated to None
        assert req.get_param_as_list('list-ish2') == ['', '0']

        # Ensure that '0' doesn't get translated to None
        assert req.get_param_as_list('list-ish3') == ['a', '', '', 'b']

        # Ensure consistency between list conventions
        assert req.get_param_as_list('thing_one') == ['1', '', '3']
        assert req.get_param_as_list('thing_one') == req.get_param_as_list('thing_two')

        store = {}
        assert req.get_param_as_list('limit', store=store) == ['1']
        assert store['limit'] == ['1']

        # Test empty elements
        assert req.get_param_as_list('empty4') == ['', '', '']
        assert req.get_param_as_list('empty5') == ['', '', '']
        assert req.get_param_as_list('empty4') == req.get_param_as_list('empty5')

    def test_list_transformer(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_qs_csv = True
        client.app.req_options.keep_blank_qs_values = False
        query_string = 'coord=1.4,13,15.1&limit=100&things=4,,1'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req

        # NOTE(kgriffs): For lists, get_param will return one of the
        # elements, but which one it will choose is undefined.
        assert req.get_param('coord') in ('1.4', '13', '15.1')

        expected = [1.4, 13.0, 15.1]
        actual = req.get_param_as_list('coord', transform=float)
        assert actual == expected

        expected = ['4', '1']
        actual = req.get_param_as_list('things', transform=str)
        assert actual == expected

        expected = [4, 1]
        actual = req.get_param_as_list('things', transform=int)
        assert actual == expected

        try:
            req.get_param_as_list('coord', transform=int)
        except Exception as ex:
            assert isinstance(ex, falcon.HTTPInvalidParam)
            assert ex.title == 'Invalid parameter'
            expected_desc = (
                'The "coord" parameter is invalid. '
                'The value is not formatted correctly.'
            )
            assert ex.description == expected_desc

    def test_param_property(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=4&bee=3&cat=2&dog=1'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        assert sorted(req.params.items()) == [
            ('ant', '4'),
            ('bee', '3'),
            ('cat', '2'),
            ('dog', '1'),
        ]

    def test_multiple_form_keys(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=1&ant=2&bee=3&cat=6&cat=5&cat=4'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        # By definition, we cannot guarantee which of the multiple keys will
        # be returned by .get_param().
        assert req.get_param('ant') in ('1', '2')
        # There is only one 'bee' key so it remains a scalar.
        assert req.get_param('bee') == '3'
        # There are three 'cat' keys; order is preserved.
        assert req.get_param('cat') in ('6', '5', '4')

    def test_multiple_keys_as_bool(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=true&ant=yes&ant=True'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_bool('ant') is True

    def test_multiple_keys_as_int(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=1&ant=2&ant=3'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_int('ant') in (1, 2, 3)

    def test_multiple_keys_as_float(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=1.1&ant=2.2&ant=3.3'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_float('ant') in (1.1, 2.2, 3.3)

    def test_multiple_form_keys_as_list(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=1&ant=2&bee=3&cat=6&cat=5&cat=4'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        # There are two 'ant' keys.
        assert req.get_param_as_list('ant') == ['1', '2']
        # There is only one 'bee' key..
        assert req.get_param_as_list('bee') == ['3']
        # There are three 'cat' keys; order is preserved.
        assert req.get_param_as_list('cat') == ['6', '5', '4']

    def test_get_date_valid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = '2015-04-20'
        query_string = 'thedate={}'.format(date_value)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_date('thedate') == date(2015, 4, 20)

    def test_get_date_missing_param(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'notthedate=2015-04-20'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_date('thedate') is None

    def test_get_date_valid_with_format(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = '20150420'
        query_string = 'thedate={}'.format(date_value)
        format_string = '%Y%m%d'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_date('thedate', format_string=format_string) == date(
            2015, 4, 20
        )

    def test_get_date_store(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = '2015-04-20'
        query_string = 'thedate={}'.format(date_value)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        store = {}
        req.get_param_as_date('thedate', store=store)
        assert len(store) != 0

    def test_get_date_invalid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = 'notarealvalue'
        query_string = 'thedate={}'.format(date_value)
        format_string = '%Y%m%d'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        with pytest.raises(HTTPInvalidParam):
            req.get_param_as_date('thedate', format_string=format_string)

    def test_get_datetime_valid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = '2015-04-20T10:10:10Z'
        query_string = 'thedate={}'.format(date_value)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_datetime('thedate') == datetime(
            2015, 4, 20, 10, 10, 10, tzinfo=timezone.utc
        )

    def test_get_datetime_missing_param(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'notthedate=2015-04-20T10:10:10Z'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_datetime('thedate') is None

    @pytest.mark.parametrize(
        'format_string, date_value, expected',
        [
            ('%Y%m%d %H:%M:%S', '20150420 10:10:10', datetime(2015, 4, 20, 10, 10, 10)),
            (
                '%Y-%m-%dT%H:%M:%SZ',
                '2015-04-20T10:10:10Z',
                datetime(2015, 4, 20, 10, 10, 10),
            ),
            (
                '%Y%m%dT%H:%M:%S.%fZ',
                '20150420T10:10:10.133701Z',
                datetime(2015, 4, 20, 10, 10, 10, 133701),
            ),
        ],
    )
    def test_get_datetime_valid_with_format(
        self, simulate_request, client, resource, format_string, date_value, expected
    ):
        client.app.add_route('/', resource)
        query_string = 'thedate={}'.format(date_value)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert (
            req.get_param_as_datetime(
                'thedate',
                format_string=format_string,
            )
            == expected
        )

    def test_get_datetime_store(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        datetime_value = '2015-04-20T10:10:10Z'
        query_string = 'thedate={}'.format(datetime_value)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        store = {}
        req.get_param_as_datetime('thedate', store=store)
        assert len(store) != 0
        assert store.get('thedate') == datetime(
            2015, 4, 20, 10, 10, 10, tzinfo=timezone.utc
        )

    def test_get_datetime_invalid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        date_value = 'notarealvalue'
        query_string = 'thedate={}'.format(date_value)
        format_string = '%Y%m%dT%H:%M:%S'
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        with pytest.raises(HTTPInvalidParam):
            req.get_param_as_datetime('thedate', format_string=format_string)

    def test_get_dict_valid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        payload_dict = {'foo': 'bar'}
        query_string = 'payload={}'.format(json.dumps(payload_dict))
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_json('payload') == payload_dict

    def test_get_dict_missing_param(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        payload_dict = {'foo': 'bar'}
        query_string = 'notthepayload={}'.format(json.dumps(payload_dict))
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_json('payload') is None

    def test_get_dict_store(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        payload_dict = {'foo': 'bar'}
        query_string = 'payload={}'.format(json.dumps(payload_dict))
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        store = {}
        req.get_param_as_json('payload', store=store)
        assert len(store) != 0

    def test_get_dict_invalid(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        payload_dict = 'foobar'
        query_string = 'payload={}'.format(payload_dict)
        simulate_request(client=client, path='/', query_string=query_string)
        req = resource.captured_req
        with pytest.raises(HTTPInvalidParam):
            req.get_param_as_json('payload')

    def test_get_param_as_json_handler_json(self, client, resource):
        client.app.add_route('/', resource)
        payload_dict = {'foo': 'bar'}
        query_string = 'payload={}'.format(json.dumps(payload_dict))
        client.app.req_options.media_handlers[falcon.MEDIA_JSON]._loads = lambda x: {
            'x': 'y'
        }
        client.simulate_get(path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_json('payload') == {'x': 'y'}

    def test_get_param_as_json_no_handler_json(self, client, resource):
        client.app.add_route('/', resource)
        payload_dict = {'foo': 'bar'}
        query_string = 'payload={}'.format(json.dumps(payload_dict))
        client.app.req_options.media_handlers.pop(falcon.MEDIA_JSON)
        client.simulate_get(path='/', query_string=query_string)
        req = resource.captured_req
        assert req.get_param_as_json('payload') == {'foo': 'bar'}

    def test_has_param(self, simulate_request, client, resource):
        client.app.add_route('/', resource)
        query_string = 'ant=1'
        simulate_request(client=client, path='/', query_string=query_string)

        req = resource.captured_req
        # There is a 'ant' key.
        assert req.has_param('ant')
        # There is not a 'bee' key..
        assert not req.has_param('bee')
        # There is not a None key
        assert not req.has_param(None)


class TestPostQueryParams:
    @pytest.mark.parametrize(
        'http_method', ('POST', 'PUT', 'PATCH', 'DELETE', 'OPTIONS')
    )
    def test_http_methods_body_expected(self, client, resource, http_method):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&limit=25'
        simulate_request_post_query_params(
            client=client, path='/', query_string=query_string, method=http_method
        )

        req = resource.captured_req
        assert req.get_param('marker') == 'deadbeef'
        assert req.get_param('limit') == '25'

    @pytest.mark.parametrize('http_method', ('GET', 'HEAD'))
    def test_http_methods_body_not_expected(self, client, resource, http_method):
        client.app.add_route('/', resource)
        query_string = 'marker=deadbeef&limit=25'
        simulate_request_post_query_params(
            client=client, path='/', query_string=query_string, method=http_method
        )

        req = resource.captured_req
        assert req.get_param('marker') is None
        assert req.get_param('limit') is None

    def test_non_ascii(self, client, resource):
        client.app.add_route('/', resource)
        value = '\u8c46\u74e3'
        query_string = b'q=' + value.encode('utf-8')
        simulate_request_post_query_params(
            client=client, path='/', query_string=query_string
        )

        req = resource.captured_req
        assert req.get_param('q') is None

    def test_empty_body(self, client, resource):
        client.app.add_route('/', resource)
        simulate_request_post_query_params(client=client, path='/', query_string=None)

        req = resource.captured_req
        assert req.get_param('q') is None

    def test_empty_body_no_content_length(self, client, resource):
        client.app.add_route('/', resource)
        simulate_request_post_query_params(client=client, path='/', query_string=None)

        req = resource.captured_req
        assert req.get_param('q') is None

    def test_explicitly_disable_auto_parse(self, client, resource):
        client.app.add_route('/', resource)
        client.app.req_options.auto_parse_form_urlencoded = False
        simulate_request_post_query_params(client=client, path='/', query_string='q=42')

        req = resource.captured_req
        assert req.get_param('q') is None

    def test_asgi_raises_error(self, util, resource):
        app = util.create_app(asgi=True)
        app.add_route('/', resource)
        with pytest.warns(deprecation.DeprecatedWarning):
            app.req_options.auto_parse_form_urlencoded = True

        with pytest.raises(RuntimeError) as exc_info:
            testing.simulate_get(app, '/')
        assert 'RequestOptions.auto_parse_form_urlencoded' in exc_info.value.args[0]


class TestPostQueryParamsDefaultBehavior:
    def test_dont_auto_parse_by_default(self, asgi, util):
        app = util.create_app(asgi)
        resource = testing.SimpleTestResource()
        app.add_route('/', resource)

        client = testing.TestClient(app)

        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        client.simulate_request(path='/', body='q=42', headers=headers)

        req = resource.captured_req
        assert req.get_param('q') is None



# Copyright 2024 by Federico Caselli
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Public Falcon type alias definitions."""

from __future__ import annotations

from typing import AsyncIterator, Dict, Optional, Protocol, TYPE_CHECKING

if TYPE_CHECKING:
    from falcon.asgi import SSEvent

__all__ = (
    'Headers',
    'ReadableIO',
    'AsyncReadableIO',
    'SSEEmitter',
)

Headers = Dict[str, str]
"""Headers dictionary returned by the framework.

.. versionadded:: 4.0
"""


# WSGI
class ReadableIO(Protocol):
    """File-like protocol that defines only a read method.

    .. versionadded:: 4.0
    """

    def read(self, n: Optional[int] = ..., /) -> bytes: ...


# ASGI
class AsyncReadableIO(Protocol):
    """Async file-like protocol that defines only a read method, and is iterable.

    .. versionadded:: 4.0
    """

    async def read(self, n: Optional[int] = ..., /) -> bytes: ...
    def __aiter__(self) -> AsyncIterator[bytes]: ...


SSEEmitter = AsyncIterator[Optional['SSEvent']]
"""Async generator or iterator over Server-Sent Events
(instances of :class:`falcon.asgi.SSEvent`).

.. versionadded:: 4.0
"""



# NOTE(myusko): A file contains temporary tests for API alias
# and it will be deleted once the API alias will be removed

import pytest

import falcon
import falcon.testing as testing
from falcon.util.deprecation import DeprecatedWarning


class CookieResource:
    def on_get(self, req, resp):
        resp.set_cookie('foo', 'bar')


@pytest.fixture
def alias_client():
    with pytest.warns(DeprecatedWarning, match='API class will be removed'):
        api = falcon.API()
    api.add_route('/get-cookie', CookieResource())
    return testing.TestClient(api)


@pytest.fixture
def app_client():
    app = falcon.App()
    app.add_route('/get-cookie', CookieResource())
    return testing.TestClient(app)


def test_cookies(alias_client, app_client):
    alias_result = alias_client.simulate_get('/get-cookie')

    alias_cookie = alias_result.cookies['foo']
    assert alias_cookie.name == 'foo'
    assert alias_cookie.value == 'bar'

    app_client_result = app_client.simulate_get('/get-cookie')
    app_cookie = app_client_result.cookies['foo']
    assert app_cookie.name == 'foo'
    assert app_cookie.value == 'bar'


def test_alias_equals_to_app(alias_client):
    with pytest.warns(DeprecatedWarning, match='API class will be removed'):
        api = falcon.API()
    assert isinstance(api, falcon.API)



import pytest

import falcon
import falcon.asgi


@pytest.fixture(params=[True, False], ids=['asgi.Response', 'Response'])
def resp_type(request):
    if request.param:
        return falcon.asgi.Response

    return falcon.Response


class TestResponseContext:
    def test_default_response_context(self, resp_type):
        resp = resp_type()

        resp.context.hello = 'World!'
        assert resp.context.hello == 'World!'
        assert resp.context['hello'] == 'World!'

        resp.context['note'] = 'Default Response.context_type used to be dict.'
        assert 'note' in resp.context
        assert hasattr(resp.context, 'note')
        assert resp.context.get('note') == resp.context['note']

    def test_custom_response_context(self, resp_type):
        class MyCustomContextType:
            pass

        class MyCustomResponse(resp_type):
            context_type = MyCustomContextType

        resp = MyCustomResponse()
        assert isinstance(resp.context, MyCustomContextType)

    def test_custom_response_context_failure(self, resp_type):
        class MyCustomResponse(resp_type):
            context_type = False

        with pytest.raises(TypeError):
            MyCustomResponse()

    def test_custom_response_context_factory(self, resp_type):
        def create_context(resp):
            return {'resp': resp}

        class MyCustomResponse(resp_type):
            context_type = create_context

        resp = MyCustomResponse()
        assert isinstance(resp.context, dict)
        assert resp.context['resp'] is resp



from datetime import datetime
from datetime import timedelta
from datetime import timezone
from http import cookies as http_cookies
import re
import warnings

import pytest

import falcon
import falcon.testing as testing
from falcon.util import http_date_to_dt
from falcon.util.deprecation import DeprecatedWarning
from falcon.util.misc import _utcnow

UNICODE_TEST_STRING = 'Unicode_\xc3\xa6\xc3\xb8'


def utcnow_naive():
    return datetime.now(timezone.utc).replace(tzinfo=None)


class CookieResource:
    def on_get(self, req, resp):
        resp.set_cookie('foo', 'bar', domain='example.com', path='/')

    def on_head(self, req, resp):
        resp.set_cookie('foo', 'bar', max_age=300)
        resp.set_cookie('bar', 'baz', http_only=False)
        resp.set_cookie('bad', 'cookie')
        resp.unset_cookie('bad')

    def on_post(self, req, resp):
        e = datetime(year=2050, month=1, day=1)  # naive
        resp.set_cookie('foo', 'bar', http_only=False, secure=False, expires=e)
        resp.unset_cookie('bad')

    def on_put(self, req, resp):
        e = datetime(
            year=2050, month=1, day=1, tzinfo=timezone(timedelta(hours=1))
        )  # aware
        resp.set_cookie('foo', 'bar', http_only=False, secure=False, expires=e)
        resp.unset_cookie('bad')


class CookieResourceMaxAgeFloatString:
    def on_get(self, req, resp):
        resp.set_cookie('foofloat', 'bar', max_age=15.3, secure=False, http_only=False)
        resp.set_cookie('foostring', 'bar', max_age='15', secure=False, http_only=False)


class CookieResourceSameSite:
    def on_get(self, req, resp):
        resp.set_cookie('foo', 'bar', same_site='Lax')
        resp.set_cookie('barz', 'barz', same_site='')

    def on_post(self, req, resp):
        resp.set_cookie('bar', 'foo', same_site='STRICT')

    def on_put(self, req, resp):
        resp.set_cookie('baz', 'foo', same_site='none')

    def on_delete(self, req, resp):
        resp.set_cookie('baz', 'foo', same_site='')


class CookieResourcePartitioned:
    def on_get(self, req, resp):
        resp.set_cookie('foo', 'bar', secure=True, partitioned=True)
        resp.set_cookie('bar', 'baz', secure=True, partitioned=False)
        resp.set_cookie('baz', 'foo', secure=True)


class CookieUnset:
    def on_get(self, req, resp):
        resp.unset_cookie('foo')
        resp.unset_cookie('bar', path='/bar')
        resp.unset_cookie('baz', domain='www.example.com')
        resp.unset_cookie('foobar', path='/foo', domain='www.example.com')
        resp.unset_cookie(
            'barfoo', same_site='none', path='/foo', domain='www.example.com'
        )


class CookieUnsetSameSite:
    def on_get(self, req, resp):
        # change lax to strict
        resp.unset_cookie('foo', same_site='Strict')
        # change strict to lax
        resp.unset_cookie('bar')
        # change none to ''
        resp.unset_cookie('baz', same_site='')
        # change '' to none
        resp.unset_cookie('barz', same_site='None')


@pytest.fixture
def client(asgi, util):
    app = util.create_app(asgi)
    app.add_route('/', CookieResource())
    app.add_route('/test-convert', CookieResourceMaxAgeFloatString())
    app.add_route('/same-site', CookieResourceSameSite())
    app.add_route('/partitioned', CookieResourcePartitioned())
    app.add_route('/unset-cookie', CookieUnset())
    app.add_route('/unset-cookie-same-site', CookieUnsetSameSite())

    return testing.TestClient(app)


# =====================================================================
# Response
# =====================================================================


def test_response_base_case(client):
    result = client.simulate_get('/')

    cookie = result.cookies['foo']

    assert cookie.name == 'foo'
    assert cookie.value == 'bar'
    assert cookie.domain == 'example.com'
    assert cookie.http_only

    # NOTE(kgriffs): Explicitly test for None to ensure
    # falcon.testing.Cookie is returning exactly what we
    # expect. Apps using falcon.testing.Cookie can be a
    # bit more cavalier if they wish.
    assert cookie.max_age is None
    assert cookie.expires is None

    assert cookie.path == '/'
    assert cookie.secure


def test_response_disable_secure_globally(client):
    client.app.resp_options.secure_cookies_by_default = False
    result = client.simulate_get('/')
    cookie = result.cookies['foo']
    assert not cookie.secure

    client.app.resp_options.secure_cookies_by_default = True
    result = client.simulate_get('/')
    cookie = result.cookies['foo']
    assert cookie.secure


def test_response_complex_case(client):
    result = client.simulate_head('/')

    assert len(result.cookies) == 3

    cookie = result.cookies['foo']
    assert cookie.value == 'bar'
    assert cookie.domain is None
    assert cookie.expires is None
    assert cookie.http_only
    assert cookie.max_age == 300
    assert cookie.path is None
    assert cookie.secure
    assert not cookie.partitioned

    cookie = result.cookies['bar']
    assert cookie.value == 'baz'
    assert cookie.domain is None
    assert cookie.expires is None
    assert not cookie.http_only
    assert cookie.max_age is None
    assert cookie.path is None
    assert cookie.secure
    assert not cookie.partitioned

    cookie = result.cookies['bad']
    assert cookie.value == ''  # An unset cookie has an empty value
    assert cookie.domain is None
    assert cookie.same_site == 'Lax'

    assert cookie.expires < _utcnow()

    # NOTE(kgriffs): I know accessing a private attr like this is
    # naughty of me, but we just need to sanity-check that the
    # string is GMT.
    assert cookie._expires.endswith('GMT')

    assert cookie.http_only
    assert cookie.max_age is None
    assert cookie.path is None
    assert cookie.secure


def test_unset_cookies(client):
    result = client.simulate_get('/unset-cookie')
    assert len(result.cookies) == 5

    def test(cookie, path, domain, same_site='Lax'):
        assert cookie.value == ''  # An unset cookie has an empty value
        assert cookie.domain == domain
        assert cookie.path == path
        assert cookie.same_site == same_site
        assert cookie.expires < _utcnow()

    test(result.cookies['foo'], path=None, domain=None)
    test(result.cookies['bar'], path='/bar', domain=None)
    test(result.cookies['baz'], path=None, domain='www.example.com')
    test(result.cookies['foobar'], path='/foo', domain='www.example.com')
    test(
        result.cookies['barfoo'],
        path='/foo',
        domain='www.example.com',
        same_site='none',
    )


def test_unset_cookies_samesite(client):
    # Test possible different samesite values in set_cookies
    # foo, bar, lax
    result_set_lax_empty = client.simulate_get('/same-site')
    # bar, foo, strict
    result_set_strict = client.simulate_post('/same-site')
    # baz, foo, none
    result_set_none = client.simulate_put('/same-site')

    def test_set(cookie, value, samesite=None):
        assert cookie.value == value
        assert cookie.same_site == samesite

    test_set(result_set_lax_empty.cookies['foo'], 'bar', samesite='Lax')
    test_set(result_set_strict.cookies['bar'], 'foo', samesite='Strict')
    test_set(result_set_none.cookies['baz'], 'foo', samesite='None')
    # barz gets set with '', that is None value
    test_set(result_set_lax_empty.cookies['barz'], 'barz')
    test_set(result_set_lax_empty.cookies['barz'], 'barz', samesite=None)

    # Unset the cookies with different samesite values
    result_unset = client.simulate_get('/unset-cookie-same-site')
    assert len(result_unset.cookies) == 4

    def test_unset(cookie, same_site='Lax'):
        assert cookie.value == ''  # An unset cookie has an empty value
        assert cookie.same_site == same_site
        assert cookie.expires < _utcnow()

    test_unset(result_unset.cookies['foo'], same_site='Strict')
    # default: bar is unset with no same_site param, so should go to Lax
    test_unset(result_unset.cookies['bar'], same_site='Lax')
    test_unset(result_unset.cookies['bar'])  # default in test_unset

    test_unset(
        result_unset.cookies['baz'], same_site=None
    )  # baz gets unset to same_site = ''
    test_unset(result_unset.cookies['barz'], same_site='None')
    # test for false
    assert result_unset.cookies['baz'].same_site != 'Strict'
    assert result_unset.cookies['foo'].same_site != 'Lax'
    assert not result_unset.cookies['baz'].same_site


def test_cookie_expires_naive(client):
    result = client.simulate_post('/')

    cookie = result.cookies['foo']
    assert cookie.value == 'bar'
    assert cookie.domain is None
    assert cookie.expires == datetime(year=2050, month=1, day=1, tzinfo=timezone.utc)
    assert not cookie.http_only
    assert cookie.max_age is None
    assert cookie.path is None
    assert not cookie.secure


def test_cookie_expires_aware(client):
    result = client.simulate_put('/')

    cookie = result.cookies['foo']
    assert cookie.value == 'bar'
    assert cookie.domain is None
    assert cookie.expires == datetime(
        year=2049, month=12, day=31, hour=23, tzinfo=timezone.utc
    )
    assert not cookie.http_only
    assert cookie.max_age is None
    assert cookie.path is None
    assert not cookie.secure


def test_cookies_setable():
    resp = falcon.Response()

    assert resp._cookies is None

    resp.set_cookie('foo', 'wrong-cookie', max_age=301)
    resp.set_cookie('foo', 'bar', max_age=300)
    resp.set_cookie('bar', 'baz', same_site='None', partitioned=True)

    morsel1 = resp._cookies['foo']
    morsel2 = resp._cookies['bar']

    assert isinstance(morsel1, http_cookies.Morsel)
    assert morsel1.key == 'foo'
    assert morsel1.value == 'bar'
    assert morsel1['max-age'] == 300

    assert isinstance(morsel2, http_cookies.Morsel)
    assert morsel2.key == 'bar'
    assert morsel2.value == 'baz'
    assert morsel2['partitioned'] is True
    assert morsel2.output() == (
        'Set-Cookie: bar=baz; HttpOnly; Partitioned; SameSite=None; Secure'
    )


@pytest.mark.parametrize('cookie_name', ('foofloat', 'foostring'))
def test_cookie_max_age_float_and_string(client, cookie_name):
    # NOTE(tbug): Falcon implicitly converts max-age values to integers,
    # to ensure RFC 6265-compliance of the attribute value.

    result = client.simulate_get('/test-convert')

    cookie = result.cookies[cookie_name]
    assert cookie.value == 'bar'
    assert cookie.domain is None
    assert cookie.expires is None
    assert not cookie.http_only
    assert cookie.max_age == 15
    assert cookie.path is None
    assert not cookie.secure


def test_response_unset_cookie():
    resp = falcon.Response()
    resp.unset_cookie('bad')
    resp.set_cookie('bad', 'cookie', max_age=300)
    resp.unset_cookie('bad')

    morsels = list(resp._cookies.values())
    assert len(morsels) == 1

    bad_cookie = morsels[0]
    assert bad_cookie['expires'] == -1

    output = bad_cookie.OutputString()
    assert 'bad=;' in output or 'bad="";' in output

    match = re.search('expires=([^;]+)', output)
    assert match

    expiration = http_date_to_dt(match.group(1), obs_date=True)
    assert expiration < _utcnow()


# =====================================================================
# Request
# =====================================================================


def test_request_cookie_parsing():
    # testing with a github-ish set of cookies
    headers = [
        (
            'Cookie',
            """
            logged_in=no;_gh_sess=eyJzZXXzaW9uX2lkIjoiN2;
            tz=Europe/Berlin; _ga =GA1.2.332347814.1422308165;
            tz2=Europe/Paris ; _ga2="line1\\012line2";
            tz3=Europe/Madrid ;_ga3= GA3.2.332347814.1422308165;
            _gat=1;
            _octo=GH1.1.201722077.1422308165
            """,
        ),
    ]

    environ = testing.create_environ(headers=headers)
    req = falcon.Request(environ)

    # NOTE(kgriffs): Test case-sensitivity
    assert req.get_cookie_values('TZ') is None
    assert 'TZ' not in req.cookies
    with pytest.raises(KeyError):
        req.cookies['TZ']

    for name, value in [
        ('logged_in', 'no'),
        ('_gh_sess', 'eyJzZXXzaW9uX2lkIjoiN2'),
        ('tz', 'Europe/Berlin'),
        ('tz2', 'Europe/Paris'),
        ('tz3', 'Europe/Madrid'),
        ('_ga', 'GA1.2.332347814.1422308165'),
        ('_ga2', 'line1\nline2'),
        ('_ga3', 'GA3.2.332347814.1422308165'),
        ('_gat', '1'),
        ('_octo', 'GH1.1.201722077.1422308165'),
    ]:
        assert name in req.cookies
        assert req.cookies[name] == value
        assert req.get_cookie_values(name) == [value]


def test_invalid_cookies_are_ignored():
    vals = [chr(i) for i in range(0x1F)]
    vals += [chr(i) for i in range(0x7F, 0xFF)]
    vals += '()<>@,;:\\"/[]?={} \x09'.split()

    for c in vals:
        headers = [
            ('Cookie', 'good_cookie=foo;bad' + c + 'cookie=bar'),
        ]

        environ = testing.create_environ(headers=headers)
        req = falcon.Request(environ)

        assert req.cookies['good_cookie'] == 'foo'
        assert 'bad' + c + 'cookie' not in req.cookies


def test_duplicate_cookie():
    headers = [
        ('Cookie', 'x=1;bad{cookie=bar; x=2;x=3 ; x=4;'),
    ]

    environ = testing.create_environ(headers=headers)
    req = falcon.Request(environ)

    assert req.cookies['x'] == '1'
    assert req.get_cookie_values('x') == ['1', '2', '3', '4']


def test_cookie_header_is_missing():
    environ = testing.create_environ(headers={})

    req = falcon.Request(environ)
    assert req.cookies == {}
    assert req.get_cookie_values('x') is None

    # NOTE(kgriffs): Test again with a new object to cover calling in the
    #   opposite order.
    req = falcon.Request(environ)
    assert req.get_cookie_values('x') is None
    assert req.cookies == {}


def test_unicode_inside_ascii_range():
    resp = falcon.Response()

    # should be ok
    resp.set_cookie('non_unicode_ascii_name_1', 'ascii_value')
    resp.set_cookie('unicode_ascii_name_1', 'ascii_value')
    resp.set_cookie('non_unicode_ascii_name_2', 'unicode_ascii_value')
    resp.set_cookie('unicode_ascii_name_2', 'unicode_ascii_value')


@pytest.mark.parametrize(
    'name', (UNICODE_TEST_STRING, UNICODE_TEST_STRING.encode('utf-8'), 42)
)
def test_non_ascii_name(name):
    resp = falcon.Response()
    with pytest.raises(KeyError):
        resp.set_cookie(name, 'ok_value')


@pytest.mark.parametrize(
    'value', (UNICODE_TEST_STRING, UNICODE_TEST_STRING.encode('utf-8'), 42)
)
def test_non_ascii_value(value):
    resp = falcon.Response()

    # NOTE(tbug): we need to grab the exception to check
    # that it is not instance of UnicodeEncodeError, so
    # we cannot simply use pytest.raises
    try:
        resp.set_cookie('ok_name', value)
    except ValueError as e:
        assert isinstance(e, ValueError)
        assert not isinstance(e, UnicodeEncodeError)
    else:
        pytest.fail('set_bad_cookie_value did not fail as expected')


def test_lax_same_site_value(client):
    result = client.simulate_get('/same-site')
    cookie = result.cookies['foo']

    assert cookie.same_site == 'Lax'


def test_strict_same_site_value(client):
    result = client.simulate_post('/same-site')
    cookie = result.cookies['bar']

    assert cookie.same_site == 'Strict'


def test_none_same_site_value(client):
    result = client.simulate_put('/same-site')
    cookie = result.cookies['baz']

    assert cookie.same_site == 'None'


def test_same_site_empty_string(client):
    result = client.simulate_delete('/same-site')
    cookie = result.cookies['baz']

    assert cookie.same_site is None


@pytest.mark.parametrize(
    'same_site', ['laX', 'lax', 'STRICT', 'strict', 'None', 'none']
)
def test_same_site_value_case_insensitive(same_site):
    resp = falcon.Response()
    resp.set_cookie('foo', 'bar', same_site=same_site)

    # NOTE(kgriffs): Verify directly, unit-test style, since we
    #   already tested end-to-end above.
    morsel = resp._cookies['foo']
    assert morsel['samesite'].lower() == same_site.lower()


@pytest.mark.parametrize('same_site', ['bogus', 'laxx', 'stric'])
def test_invalid_same_site_value(same_site):
    resp = falcon.Response()

    with pytest.raises(ValueError):
        resp.set_cookie('foo', 'bar', same_site=same_site)


def test_partitioned_value(client):
    result = client.simulate_get('/partitioned')

    cookie = result.cookies['foo']
    assert cookie.partitioned

    cookie = result.cookies['bar']
    assert not cookie.partitioned

    cookie = result.cookies['baz']
    assert not cookie.partitioned


def test_unset_cookie_deprecation_warning():
    resp = falcon.Response()

    # Test that using the deprecated 'samesite' parameter raises a warning
    with pytest.warns(
        DeprecatedWarning, match='The "samesite" parameter is deprecated'
    ):
        resp.unset_cookie('test', samesite='Strict')

    # Verify the cookie was still set correctly with the deprecated parameter
    morsel = resp._cookies['test']
    assert morsel['samesite'] == 'Strict'

    # Test that using the new 'same_site' parameter doesn't raise a warning
    with warnings.catch_warnings():
        warnings.simplefilter('error')
        resp.unset_cookie('test2', same_site='Lax')

    # Verify the cookie was set correctly with the new parameter
    morsel2 = resp._cookies['test2']
    assert morsel2['samesite'] == 'Lax'

    # Test that when both parameters are provided, deprecated one is used with warning
    with pytest.warns(
        DeprecatedWarning, match='The "samesite" parameter is deprecated'
    ):
        resp.unset_cookie('test3', samesite='None', same_site='Strict')

    # Verify the deprecated parameter value was used
    morsel3 = resp._cookies['test3']
    assert morsel3['samesite'] == 'None'



# Copyright 2015 by Hurricane Labs LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HTTPStatus exception class."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING

from falcon.util import http_status_to_code

if TYPE_CHECKING:
    from falcon._typing import HeaderArg
    from falcon._typing import ResponseStatus


class HTTPStatus(Exception):
    """Represents a generic HTTP status.

    Raise an instance of this class from a hook, middleware, or
    responder to short-circuit request processing in a manner similar
    to ``falcon.HTTPError``, but for non-error status codes.

    Args:
        status (Union[str,int]): HTTP status code or line (e.g.,
            ``'400 Bad Request'``). This may be set to a member of
            :class:`http.HTTPStatus`, an HTTP status line string or byte
            string (e.g., ``'200 OK'``), or an ``int``.
        headers (dict): Extra headers to add to the response.
        text (str): String representing response content. Falcon will encode
            this value as UTF-8 in the response.
    """

    __slots__ = ('status', 'headers', 'text')

    status: ResponseStatus
    """The HTTP status line or integer code for the status that this exception
    represents.
    """
    headers: Optional[HeaderArg]
    """Extra headers to add to the response."""
    text: Optional[str]
    """String representing response content.
    Falcon will encode this value as UTF-8 in the response.
    """

    def __init__(
        self,
        status: ResponseStatus,
        headers: Optional[HeaderArg] = None,
        text: Optional[str] = None,
    ) -> None:
        self.status = status
        self.headers = headers
        self.text = text

    @property
    def status_code(self) -> int:
        """HTTP status code normalized from :attr:`status`."""
        return http_status_to_code(self.status)



import hashlib
import io
import os.path

import falcon

HERE = os.path.abspath(os.path.dirname(__file__))


class Forms:
    def on_post(self, req, resp):
        parts = {}

        for part in req.media:
            # NOTE(vytas): SHA1 is no longer recommended for cryptographic
            #   purposes, but here we are only using it for integrity checking.
            sha1 = hashlib.sha1()
            while True:
                chunk = part.stream.read(io.DEFAULT_BUFFER_SIZE)
                if not chunk:
                    break

                sha1.update(chunk)

            parts[part.name] = {
                'filename': part.filename,
                'sha1': sha1.hexdigest(),
            }

        resp.media = parts


class Hello:
    def on_get(self, req, resp):
        resp.set_header('X-Falcon', 'peregrine')

        resp.content_type = falcon.MEDIA_TEXT
        resp.text = 'Hello, World!\n'

    def on_get_deprecated(self, req, resp):
        resp.set_header('X-Falcon', 'deprecated')

        resp.content_type = falcon.MEDIA_TEXT
        resp.text = 'Hello, World!\n'

        resp.add_link('/removed-methods/add-link', 'bookmark')


app = application = falcon.App()
app.add_route('/forms', Forms())
app.add_route('/hello', Hello())
app.add_route('/deprecated', Hello(), suffix='deprecated')
app.add_static_route('/tests', HERE, downloadable=True)



import datetime
import itertools

import pytest

import falcon
from falcon.request import Request
from falcon.request import RequestOptions
from falcon.request_helpers import _parse_etags
import falcon.testing as testing
import falcon.uri
from falcon.util import DeprecatedWarning
from falcon.util.structures import ETag

_HTTP_VERSIONS = ['1.0', '1.1', '2']


def _make_etag(value, is_weak=False):
    """Creates and returns an ETag object.

    Args:
        value (str): Unquated entity tag value
        is_weak (bool): The weakness indicator

    Returns:
        A ``str``-like Etag instance with weakness indicator.

    """
    etag = ETag(value)

    etag.is_weak = is_weak
    return etag


# NOTE(vytas): create_req is very heavily used in this module in unittest-style
#   classes, so we simply recreate the function here.
def create_req(asgi, options=None, **environ_or_scope_kwargs):
    create_method = testing.create_asgi_req if asgi else testing.create_req
    return create_method(options=options, **environ_or_scope_kwargs)


def test_missing_qs():
    env = testing.create_environ()
    if 'QUERY_STRING' in env:
        del env['QUERY_STRING']

    # Should not cause an exception when Request is instantiated
    Request(env)


def test_app_missing():
    env = testing.create_environ()
    del env['SCRIPT_NAME']
    req = Request(env)

    assert req.root_path == ''
    with pytest.warns(DeprecatedWarning):
        assert req.app == ''


class TestRequestAttributes:
    def setup_method(self, method):
        asgi = self._item.callspec.getparam('asgi')

        self.qs = 'marker=deadbeef&limit=10'

        self.headers = {
            'Content-Type': 'text/plain',
            'Content-Length': '4829',
            'Authorization': '',
        }

        self.root_path = '/test'
        self.path = '/hello'
        self.relative_uri = self.path + '?' + self.qs

        self.req = create_req(
            asgi,
            root_path=self.root_path,
            port=8080,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        self.req_noqs = create_req(
            asgi, root_path=self.root_path, path='/hello', headers=self.headers
        )

    def test_empty(self, asgi):
        assert self.req.auth is None

    def test_host(self, asgi):
        assert self.req.host == testing.DEFAULT_HOST

    def test_subdomain(self, asgi):
        req = create_req(asgi, host='com', path='/hello', headers=self.headers)
        assert req.subdomain is None

        req = create_req(asgi, host='example.com', path='/hello', headers=self.headers)
        assert req.subdomain == 'example'

        req = create_req(
            asgi, host='highwire.example.com', path='/hello', headers=self.headers
        )
        assert req.subdomain == 'highwire'

        req = create_req(
            asgi,
            host='lb01.dfw01.example.com',
            port=8080,
            path='/hello',
            headers=self.headers,
        )
        assert req.subdomain == 'lb01'

        # NOTE(kgriffs): Behavior for IP addresses is undefined,
        # so just make sure it doesn't blow up.
        req = create_req(asgi, host='127.0.0.1', path='/hello', headers=self.headers)
        assert type(req.subdomain) is str

        # NOTE(kgriffs): Test fallback to SERVER_NAME by using
        # HTTP 1.0, which will cause .create_environ to not set
        # HTTP_HOST.
        req = create_req(
            asgi,
            http_version='1.0',
            host='example.com',
            path='/hello',
            headers=self.headers,
        )
        assert req.subdomain == 'example'

    def test_reconstruct_url(self, asgi):
        req = self.req

        scheme = req.scheme
        host = req.get_header('host')
        app = req.root_path
        with pytest.warns(DeprecatedWarning):
            assert req.app == app
        path = req.path
        query_string = req.query_string

        expected_prefix = ''.join([scheme, '://', host, app])
        expected_uri = ''.join([expected_prefix, path, '?', query_string])

        assert req.uri == expected_uri
        assert req.prefix == expected_prefix
        assert req.prefix == expected_prefix  # Check cached value

    @pytest.mark.parametrize(
        'test_path',
        [
            '/hello_\u043f\u0440\u0438\u0432\u0435\u0442',
            '/test/%E5%BB%B6%E5%AE%89',
            '/test/%C3%A4%C3%B6%C3%BC%C3%9F%E2%82%AC',
        ],
    )
    def test_nonlatin_path(self, asgi, test_path):
        # NOTE(kgriffs): When a request comes in, web servers decode
        # the path.  The decoded path may contain UTF-8 characters,
        # but according to the WSGI spec, no strings can contain chars
        # outside ISO-8859-1. Therefore, to reconcile the URI
        # encoding standard that allows UTF-8 with the WSGI spec
        # that does not, WSGI servers tunnel the string via
        # ISO-8859-1. falcon.testing.create_environ() mimics this
        # behavior, e.g.:
        #
        #   tunnelled_path = path.encode('utf-8').decode('iso-8859-1')
        #
        # falcon.Request does the following to reverse the process:
        #
        #   path = tunnelled_path.encode('iso-8859-1').decode('utf-8', 'replace')
        #

        req = create_req(asgi, host='com', path=test_path, headers=self.headers)

        assert req.path == falcon.uri.decode(test_path)

    def test_uri(self, asgi):
        prefix = 'http://' + testing.DEFAULT_HOST + ':8080' + self.root_path
        uri = prefix + self.relative_uri

        assert self.req.url == uri
        assert self.req.prefix == prefix

        # NOTE(kgriffs): Call twice to check caching works
        assert self.req.uri == uri
        assert self.req.uri == uri

        uri_noqs = 'http://' + testing.DEFAULT_HOST + self.root_path + self.path
        assert self.req_noqs.uri == uri_noqs

    def test_uri_https(self, asgi):
        # =======================================================
        # Default port, implicit
        # =======================================================
        req = create_req(asgi, path='/hello', scheme='https')
        uri = 'https://' + testing.DEFAULT_HOST + '/hello'

        assert req.uri == uri

        # =======================================================
        # Default port, explicit
        # =======================================================
        req = create_req(asgi, path='/hello', scheme='https', port=443)
        uri = 'https://' + testing.DEFAULT_HOST + '/hello'

        assert req.uri == uri

        # =======================================================
        # Non-default port
        # =======================================================
        req = create_req(asgi, path='/hello', scheme='https', port=22)
        uri = 'https://' + testing.DEFAULT_HOST + ':22/hello'

        assert req.uri == uri

    def test_uri_http_1_0(self, asgi):
        # =======================================================
        # HTTP, 80
        # =======================================================
        req = create_req(
            asgi,
            http_version='1.0',
            root_path=self.root_path,
            port=80,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        uri = 'http://' + testing.DEFAULT_HOST + self.root_path + self.relative_uri

        assert req.uri == uri

        # =======================================================
        # HTTP, 80
        # =======================================================
        req = create_req(
            asgi,
            http_version='1.0',
            root_path=self.root_path,
            port=8080,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        uri = (
            'http://'
            + testing.DEFAULT_HOST
            + ':8080'
            + self.root_path
            + self.relative_uri
        )

        assert req.uri == uri

        # =======================================================
        # HTTP, 80
        # =======================================================
        req = create_req(
            asgi,
            http_version='1.0',
            scheme='https',
            root_path=self.root_path,
            port=443,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        uri = 'https://' + testing.DEFAULT_HOST + self.root_path + self.relative_uri

        assert req.uri == uri

        # =======================================================
        # HTTP, 80
        # =======================================================
        req = create_req(
            asgi,
            http_version='1.0',
            scheme='https',
            root_path=self.root_path,
            port=22,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        uri = (
            'https://'
            + testing.DEFAULT_HOST
            + ':22'
            + self.root_path
            + self.relative_uri
        )

        assert req.uri == uri

    def test_relative_uri(self, asgi):
        assert self.req.relative_uri == self.root_path + self.relative_uri
        assert self.req_noqs.relative_uri == self.root_path + self.path

        req_noapp = create_req(
            asgi, path='/hello', query_string=self.qs, headers=self.headers
        )

        assert req_noapp.relative_uri == self.relative_uri

        req_noapp = create_req(
            asgi, path='/hello/', query_string=self.qs, headers=self.headers
        )

        relative_trailing_uri = self.path + '/?' + self.qs
        # NOTE(kgriffs): Call twice to check caching works
        assert req_noapp.relative_uri == relative_trailing_uri
        assert req_noapp.relative_uri == relative_trailing_uri

        options = RequestOptions()
        options.strip_url_path_trailing_slash = False
        req_noapp = create_req(
            asgi,
            options=options,
            path='/hello/',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req_noapp.relative_uri == '/hello/' + '?' + self.qs

    def test_client_accepts(self, asgi):
        headers = {'Accept': 'application/xml'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('application/xml')

        headers = {'Accept': '*/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('application/xml')
        assert req.client_accepts('application/json')
        assert req.client_accepts('application/x-msgpack')

        headers = {'Accept': 'application/x-msgpack'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts('application/xml')
        assert not req.client_accepts('application/json')
        assert req.client_accepts('application/x-msgpack')

        headers = {}  # NOTE(kgriffs): Equivalent to '*/*' per RFC
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('application/xml')

        headers = {'Accept': 'application/json'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts('application/xml')

        headers = {'Accept': 'application/x-msgpack'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('application/x-msgpack')

        headers = {'Accept': 'application/xm'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts('application/xml')

        headers = {'Accept': 'application/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('application/json')
        assert req.client_accepts('application/xml')
        assert req.client_accepts('application/x-msgpack')

        headers = {'Accept': 'text/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('text/plain')
        assert req.client_accepts('text/csv')
        assert not req.client_accepts('application/xhtml+xml')

        headers = {'Accept': 'text/*, application/xhtml+xml; q=0.0'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('text/plain')
        assert req.client_accepts('text/csv')
        assert not req.client_accepts('application/xhtml+xml')

        headers = {'Accept': 'text/*; q=0.1, application/xhtml+xml; q=0.5'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('text/plain')
        assert req.client_accepts('application/xhtml+xml')

        headers = {'Accept': 'text/*,         application/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('text/plain')
        assert req.client_accepts('application/xml')
        assert req.client_accepts('application/json')
        assert req.client_accepts('application/x-msgpack')

        headers = {'Accept': 'text/*,application/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts('text/plain')
        assert req.client_accepts('application/xml')
        assert req.client_accepts('application/json')
        assert req.client_accepts('application/x-msgpack')

    def test_client_accepts_bogus(self, asgi):
        headers = {'Accept': '~'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts('text/plain')
        assert not req.client_accepts('application/json')

    def test_client_accepts_props(self, asgi):
        headers = {'Accept': 'application/xml'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts_xml
        assert not req.client_accepts_json
        assert not req.client_accepts_msgpack

        headers = {'Accept': 'application/*'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts_xml
        assert req.client_accepts_json
        assert req.client_accepts_msgpack

        headers = {'Accept': 'application/json'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts_xml
        assert req.client_accepts_json
        assert not req.client_accepts_msgpack

        headers = {'Accept': 'application/x-msgpack'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts_xml
        assert not req.client_accepts_json
        assert req.client_accepts_msgpack

        headers = {'Accept': 'application/msgpack'}
        req = create_req(asgi, headers=headers)
        assert not req.client_accepts_xml
        assert not req.client_accepts_json
        assert req.client_accepts_msgpack

        headers = {'Accept': 'application/json,application/xml,application/x-msgpack'}
        req = create_req(asgi, headers=headers)
        assert req.client_accepts_xml
        assert req.client_accepts_json
        assert req.client_accepts_msgpack

    def test_client_prefers(self, asgi):
        headers = {'Accept': 'application/xml'}
        req = create_req(asgi, headers=headers)
        preferred_type = req.client_prefers(['application/xml'])
        assert preferred_type == 'application/xml'

        headers = {'Accept': '*/*'}
        preferred_type = req.client_prefers(('application/xml', 'application/json'))

        # NOTE(kgriffs): If client doesn't care, "prefer" the first one
        assert preferred_type == 'application/xml'

        headers = {'Accept': 'text/*; q=0.1, application/xhtml+xml; q=0.5'}
        req = create_req(asgi, headers=headers)
        preferred_type = req.client_prefers(['application/xhtml+xml'])
        assert preferred_type == 'application/xhtml+xml'

        headers = {'Accept': '3p12845j;;;asfd;'}
        req = create_req(asgi, headers=headers)
        preferred_type = req.client_prefers(['application/xhtml+xml'])
        assert preferred_type is None

    def test_range(self, asgi):
        headers = {'Range': 'bytes=10-'}
        req = create_req(asgi, headers=headers)
        assert req.range == (10, -1)

        headers = {'Range': 'bytes=10-20'}
        req = create_req(asgi, headers=headers)
        assert req.range == (10, 20)

        headers = {'Range': 'bytes=-10240'}
        req = create_req(asgi, headers=headers)
        assert req.range == (-10240, -1)

        headers = {'Range': 'bytes=0-2'}
        req = create_req(asgi, headers=headers)
        assert req.range == (0, 2)

        headers = {'Range': ''}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPInvalidHeader):
            req.range

        req = create_req(asgi)
        assert req.range is None

    def test_range_unit(self, asgi):
        headers = {'Range': 'bytes=10-'}
        req = create_req(asgi, headers=headers)
        assert req.range == (10, -1)
        assert req.range_unit == 'bytes'

        headers = {'Range': 'items=10-'}
        req = create_req(asgi, headers=headers)
        assert req.range == (10, -1)
        assert req.range_unit == 'items'

        headers = {'Range': ''}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPInvalidHeader):
            req.range_unit

        req = create_req(asgi)
        assert req.range_unit is None

    def test_range_invalid(self, asgi):
        headers = {'Range': 'bytes=10240'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=-'}
        expected_desc = (
            'The value provided for the "Range" header is '
            'invalid. The range offsets are missing.'
        )
        self._test_error_details(
            headers,
            'range',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

        headers = {'Range': 'bytes=--'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=--1'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=--0'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=-3-'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=-3-4'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=4-3'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=3-3-4'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=3-3-'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=3-3- '}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=fizbit'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=a-'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=a-3'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=-b'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=3-b'}
        req = create_req(asgi, headers=headers)
        with pytest.raises(falcon.HTTPBadRequest):
            req.range

        headers = {'Range': 'bytes=x-y'}
        expected_desc = (
            'The value provided for the "Range" header is '
            'invalid. It must be a range formatted '
            'according to RFC 7233.'
        )
        self._test_error_details(
            headers,
            'range',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

        headers = {'Range': 'bytes=0-0,-1'}
        expected_desc = (
            'The value provided for the "Range" '
            'header is invalid. The value must be a '
            'continuous range.'
        )
        self._test_error_details(
            headers,
            'range',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

        headers = {'Range': '10-'}
        expected_desc = (
            'The value provided for the "Range" '
            'header is invalid. The value must be '
            "prefixed with a range unit, e.g. 'bytes='"
        )
        self._test_error_details(
            headers,
            'range',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

    def test_missing_attribute_header(self, asgi):
        req = create_req(asgi)
        assert req.range is None

        req = create_req(asgi)
        assert req.content_length is None

    def test_content_length(self, asgi):
        headers = {'content-length': '5656'}
        req = create_req(asgi, headers=headers)
        assert req.content_length == 5656

        headers = {'content-length': ''}
        req = create_req(asgi, headers=headers)
        assert req.content_length is None

    def test_bogus_content_length_nan(self, asgi):
        headers = {'content-length': 'fuzzy-bunnies'}
        expected_desc = (
            'The value provided for the '
            '"Content-Length" header is invalid. The value '
            'of the header must be a number.'
        )
        self._test_error_details(
            headers,
            'content_length',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

    def test_bogus_content_length_neg(self, asgi):
        headers = {'content-length': '-1'}
        expected_desc = (
            'The value provided for the "Content-Length" '
            'header is invalid. The value of the header '
            'must be a positive number.'
        )
        self._test_error_details(
            headers,
            'content_length',
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc,
            asgi,
        )

    @pytest.mark.parametrize(
        'header,attr',
        [
            ('Date', 'date'),
            ('If-Modified-Since', 'if_modified_since'),
            ('If-Unmodified-Since', 'if_unmodified_since'),
        ],
    )
    def test_date(self, asgi, header, attr):
        date = datetime.datetime(2013, 4, 4, 5, 19, 18, tzinfo=datetime.timezone.utc)
        date_str = 'Thu, 04 Apr 2013 05:19:18 GMT'

        headers = {header: date_str}
        req = create_req(asgi, headers=headers)
        assert getattr(req, attr) == date

    @pytest.mark.parametrize(
        'header,attr',
        [
            ('Date', 'date'),
            ('If-Modified-Since', 'if_modified_since'),
            ('If-Unmodified-Since', 'if_unmodified_since'),
        ],
    )
    def test_date_invalid(self, asgi, header, attr):
        # Date formats don't conform to RFC 1123
        headers = {header: 'Thu, 04 Apr 2013'}
        expected_desc = (
            'The value provided for the "{}" '
            'header is invalid. It must be formatted '
            'according to RFC 7231, Section 7.1.1.1'
        )

        self._test_error_details(
            headers,
            attr,
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc.format(header),
            asgi,
        )

        headers = {header: ''}
        self._test_error_details(
            headers,
            attr,
            falcon.HTTPInvalidHeader,
            'Invalid header value',
            expected_desc.format(header),
            asgi,
        )

    @pytest.mark.parametrize(
        'attr', ('date', 'if_modified_since', 'if_unmodified_since')
    )
    def test_date_missing(self, asgi, attr):
        req = create_req(asgi)
        assert getattr(req, attr) is None

    @pytest.mark.parametrize(
        'name,value,attr,default',
        [
            ('Accept', 'x-falcon', 'accept', '*/*'),
            (
                'Authorization',
                'HMAC_SHA1 c590afa9bb59191ffab30f223791e82d3fd3e3af',
                'auth',
                None,
            ),
            ('Content-Type', 'text/plain', 'content_type', None),
            ('Expect', '100-continue', 'expect', None),
            ('If-Range', 'Wed, 21 Oct 2015 07:28:00 GMT', 'if_range', None),
            (
                'User-Agent',
                'testing/3.0',
                'user_agent',
                'falcon-client/' + falcon.__version__,
            ),
            ('Referer', 'https://www.google.com/', 'referer', None),
        ],
    )
    def test_attribute_headers(self, asgi, name, value, attr, default):
        headers = {name: value}
        req = create_req(asgi, headers=headers)
        assert getattr(req, attr) == value

        req = create_req(asgi)
        assert getattr(req, attr) == default

    def test_method(self, asgi):
        assert self.req.method == 'GET'

        self.req = create_req(asgi, path='', method='HEAD')
        assert self.req.method == 'HEAD'

    def test_empty_path(self, asgi):
        self.req = create_req(asgi, path='')
        assert self.req.path == '/'

    def test_content_type_method(self, asgi):
        assert self.req.get_header('content-type') == 'text/plain'

    def test_content_length_method(self, asgi):
        assert self.req.get_header('content-length') == '4829'

    # TODO(kgriffs): Migrate to pytest and parametrized fixtures
    # to DRY things up a bit.
    @pytest.mark.parametrize('http_version', _HTTP_VERSIONS)
    def test_port_explicit(self, asgi, http_version):
        port = 9000
        req = create_req(
            asgi,
            http_version=http_version,
            port=port,
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req.port == port

    @pytest.mark.parametrize('http_version', _HTTP_VERSIONS)
    def test_scheme_https(self, asgi, http_version):
        scheme = 'https'
        req = create_req(
            asgi,
            http_version=http_version,
            scheme=scheme,
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req.scheme == scheme
        assert req.port == 443

    @pytest.mark.parametrize(
        'http_version, set_forwarded_proto',
        list(itertools.product(_HTTP_VERSIONS, [True, False])),
    )
    def test_scheme_http(self, asgi, http_version, set_forwarded_proto):
        scheme = 'http'
        forwarded_scheme = 'HttPs'

        headers = dict(self.headers)

        if set_forwarded_proto:
            headers['X-Forwarded-Proto'] = forwarded_scheme

        req = create_req(
            asgi,
            http_version=http_version,
            scheme=scheme,
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=headers,
        )

        assert req.scheme == scheme
        assert req.port == 80

        if set_forwarded_proto:
            assert req.forwarded_scheme == forwarded_scheme.lower()
        else:
            assert req.forwarded_scheme == scheme

    @pytest.mark.parametrize('http_version', _HTTP_VERSIONS)
    def test_netloc_default_port(self, asgi, http_version):
        req = create_req(
            asgi,
            http_version=http_version,
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req.netloc == 'falconframework.org'

    @pytest.mark.parametrize('http_version', _HTTP_VERSIONS)
    def test_netloc_nondefault_port(self, asgi, http_version):
        req = create_req(
            asgi,
            http_version=http_version,
            port='8080',
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req.netloc == 'falconframework.org:8080'

    @pytest.mark.parametrize('http_version', _HTTP_VERSIONS)
    def test_netloc_from_env(self, asgi, http_version):
        port = 9000
        host = 'example.org'

        req = create_req(
            asgi,
            http_version=http_version,
            host=host,
            port=port,
            root_path=self.root_path,
            path='/hello',
            query_string=self.qs,
            headers=self.headers,
        )

        assert req.port == port
        assert req.netloc == '{}:{}'.format(host, port)

    def test_app_present(self, asgi):
        req = create_req(asgi, root_path='/moving-pictures')
        with pytest.warns(DeprecatedWarning):
            assert req.app == '/moving-pictures'
        assert req.root_path == '/moving-pictures'

    def test_app_blank(self, asgi):
        req = create_req(asgi, root_path='')
        with pytest.warns(DeprecatedWarning):
            assert req.app == ''
        assert req.root_path == ''

    @pytest.mark.parametrize(
        'etag,expected_value',
        [
            ('', None),
            (' ', None),
            ('   ', None),
            ('\t', None),
            (' \t', None),
            (',', None),
            (',,', None),
            (',, ', None),
            (', , ', None),
            ('*', ['*']),
            ('W/"67ab43"', [_make_etag('67ab43', is_weak=True)]),
            ('w/"67ab43"', [_make_etag('67ab43', is_weak=True)]),
            (' w/"67ab43"', [_make_etag('67ab43', is_weak=True)]),
            ('w/"67ab43" ', [_make_etag('67ab43', is_weak=True)]),
            ('w/"67ab43 " ', [_make_etag('67ab43 ', is_weak=True)]),
            ('"67ab43"', [_make_etag('67ab43')]),
            (' "67ab43"', [_make_etag('67ab43')]),
            (' "67ab43" ', [_make_etag('67ab43')]),
            ('"67ab43" ', [_make_etag('67ab43')]),
            ('" 67ab43" ', [_make_etag(' 67ab43')]),
            ('67ab43"', [_make_etag('67ab43"')]),
            ('"67ab43', [_make_etag('"67ab43')]),
            ('67ab43', [_make_etag('67ab43')]),
            ('67ab43 ', [_make_etag('67ab43')]),
            ('  67ab43 ', [_make_etag('67ab43')]),
            ('  67ab43', [_make_etag('67ab43')]),
            (
                # NOTE(kgriffs): To simplify parsing and improve performance, we
                #   do not attempt to handle unquoted entity-tags when there is
                #   a list; it is non-standard anyway, and has been since 1999.
                'W/"67ab43", "54ed21", junk"F9,22", junk "41, 7F", '
                'unquoted, w/"22, 41, 7F", "", W/""',
                [
                    _make_etag('67ab43', is_weak=True),
                    _make_etag('54ed21'),
                    # NOTE(kgriffs): Test that the ETag initializer defaults to
                    #   is_weak == False
                    ETag('F9,22'),
                    _make_etag('41, 7F'),
                    _make_etag('22, 41, 7F', is_weak=True),
                    # NOTE(kgriffs): According to the grammar in RFC 7232, zero
                    #  etagc's is acceptable.
                    _make_etag(''),
                    _make_etag('', is_weak=True),
                ],
            ),
        ],
    )
    @pytest.mark.parametrize(
        'name,attr',
        [
            ('If-Match', 'if_match'),
            ('If-None-Match', 'if_none_match'),
        ],
    )
    def test_etag(self, asgi, name, attr, etag, expected_value):
        headers = {name: etag}
        req = create_req(asgi, headers=headers)

        # NOTE(kgriffs): Loop in order to test caching
        for __ in range(3):
            value = getattr(req, attr)

            if expected_value is None:
                assert value is None
                return

            assert value is not None

            for element, expected_element in zip(value, expected_value):
                assert element == expected_element
                if isinstance(expected_element, ETag):
                    assert element.is_weak == expected_element.is_weak

    def test_etag_is_missing(self, asgi):
        # NOTE(kgriffs): Loop in order to test caching
        for __ in range(3):
            assert self.req.if_match is None
            assert self.req.if_none_match is None

    @pytest.mark.parametrize('header_value', ['', ' ', '  '])
    def test_etag_parsing_helper(self, asgi, header_value):
        # NOTE(kgriffs): Test a couple of cases that are not directly covered
        #   elsewhere (but that we want the helper to still support
        #   for the sake of avoiding surprises if they are ever called without
        #   preflighting the header value).

        assert _parse_etags(header_value) is None

    # -------------------------------------------------------------------------
    # Helpers
    # -------------------------------------------------------------------------

    def _test_error_details(
        self, headers, attr_name, error_type, title, description, asgi
    ):
        req = create_req(asgi, headers=headers)

        try:
            getattr(req, attr_name)
            pytest.fail('{} not raised'.format(error_type.__name__))
        except error_type as ex:
            assert ex.title == title
            assert ex.description == description



import re

from falcon.asgi import Request
from falcon.asgi import WebSocket

from .hub import Hub


class Chat:
    ALL = re.compile(r'^/all\s+(.+)$')
    MSG = re.compile(r'^/msg\s+(\w+)\s+(.+)$')

    def __init__(self, hub: Hub):
        self._hub = hub

    async def on_websocket(self, req: Request, ws: WebSocket, name: str) -> None:
        await ws.accept()

        try:
            await self._hub.broadcast(f'{name} CONNECTED')
            self._hub.add_user(name, ws)

            await ws.send_text(f'Hello, {name}!')

            while True:
                message = await ws.receive_text()

                if message == '/quit':
                    await ws.send_text(f'Bye, {name}!')
                    await ws.close(4001, 'quit command')
                    break

                command = self.ALL.match(message)
                if command:
                    text = command.group(1)
                    await self._hub.broadcast(f'[{name}] {text}')
                    continue

                command = self.MSG.match(message)
                if command:
                    recipient, text = command.groups()
                    await self._hub.message(recipient, f'[{name}] {text}')
                    continue

                await ws.send_text('Supported commands: /all /msg /quit')

        finally:
            self._hub.remove_user(name)
            await self._hub.broadcast(f'{name} DISCONNECTED')



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hook decorators."""

from __future__ import annotations

from functools import wraps
from inspect import getmembers
from inspect import iscoroutinefunction
import re
from typing import (
    Any,
    Awaitable,
    Callable,
    cast,
    Dict,
    List,
    Protocol,
    Tuple,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from falcon.constants import COMBINED_METHODS
from falcon.util.misc import get_argnames
from falcon.util.sync import _wrap_non_coroutine_unsafe

if TYPE_CHECKING:
    import falcon as wsgi
    from falcon import asgi
    from falcon._typing import AsgiResponderMethod
    from falcon._typing import Resource
    from falcon._typing import Responder
    from falcon._typing import ResponderMethod


# TODO: when targeting only 3.10+ these protocol would no longer be needed, since
# ParamSpec could be used together with Concatenate to use a simple Callable
# to type the before and after functions. This approach was prototyped in
# https://github.com/falconry/falcon/pull/2234
class SyncBeforeFn(Protocol):
    def __call__(
        self,
        req: wsgi.Request,
        resp: wsgi.Response,
        resource: Resource,
        params: Dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None: ...


class AsyncBeforeFn(Protocol):
    def __call__(
        self,
        req: asgi.Request,
        resp: asgi.Response,
        resource: Resource,
        params: Dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[None]: ...


BeforeFn = Union[SyncBeforeFn, AsyncBeforeFn]


class SyncAfterFn(Protocol):
    def __call__(
        self,
        req: wsgi.Request,
        resp: wsgi.Response,
        resource: Resource,
        *args: Any,
        **kwargs: Any,
    ) -> None: ...


class AsyncAfterFn(Protocol):
    def __call__(
        self,
        req: asgi.Request,
        resp: asgi.Response,
        resource: Resource,
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[None]: ...


AfterFn = Union[SyncAfterFn, AsyncAfterFn]
_R = TypeVar('_R', bound=Union['Responder', 'Resource'])


_DECORABLE_METHOD_NAME = re.compile(
    r'^on_({})(_\w+)?$'.format('|'.join(method.lower() for method in COMBINED_METHODS))
)


def before(action: BeforeFn, *args: Any, **kwargs: Any) -> Callable[[_R], _R]:
    """Execute the given action function *before* the responder.

    The `params` argument that is passed to the hook
    contains only the fields from the URI template path; it does not
    include query string values.

    Hooks may inject extra params as needed. For example::

        def do_something(req, resp, resource, params):
            try:
                params['id'] = int(params['id'])
            except ValueError:
                raise falcon.HTTPBadRequest(title='Invalid ID',
                                            description='ID was not valid.')

            params['answer'] = 42

    Args:
        action (callable): A function of the form
            ``func(req, resp, resource, params)``, where `resource` is a
            reference to the resource class instance associated with the
            request and `params` is a dict of URI template field names,
            if any, that will be passed into the resource responder as
            kwargs.

        *args: Any additional arguments will be passed to *action* in the
            order given, immediately following the *req*, *resp*, *resource*,
            and *params* arguments.

    Keyword Args:
        **kwargs: Any additional keyword arguments will be passed through to
            *action*.
    """

    def _before(responder_or_resource: _R) -> _R:
        if isinstance(responder_or_resource, type):
            for responder_name, responder in getmembers(
                responder_or_resource, callable
            ):
                if _DECORABLE_METHOD_NAME.match(responder_name):
                    responder = cast('Responder', responder)
                    do_before_all = _wrap_with_before(responder, action, args, kwargs)

                    setattr(responder_or_resource, responder_name, do_before_all)

            return cast(_R, responder_or_resource)

        else:
            responder = cast('Responder', responder_or_resource)
            do_before_one = _wrap_with_before(responder, action, args, kwargs)

            return cast(_R, do_before_one)

    return _before


def after(action: AfterFn, *args: Any, **kwargs: Any) -> Callable[[_R], _R]:
    """Execute the given action function *after* the responder.

    Args:
        action (callable): A function of the form
            ``func(req, resp, resource)``, where `resource` is a
            reference to the resource class instance associated with the
            request

        *args: Any additional arguments will be passed to *action* in the
            order given, immediately following the *req*, *resp* and *resource*
            arguments.

    Keyword Args:
        **kwargs: Any additional keyword arguments will be passed through to
            *action*.
    """

    def _after(responder_or_resource: _R) -> _R:
        if isinstance(responder_or_resource, type):
            for responder_name, responder in getmembers(
                responder_or_resource, callable
            ):
                if _DECORABLE_METHOD_NAME.match(responder_name):
                    responder = cast('Responder', responder)
                    do_after_all = _wrap_with_after(responder, action, args, kwargs)

                    setattr(responder_or_resource, responder_name, do_after_all)

            return cast(_R, responder_or_resource)

        else:
            responder = cast('Responder', responder_or_resource)
            do_after_one = _wrap_with_after(responder, action, args, kwargs)

            return cast(_R, do_after_one)

    return _after


# -----------------------------------------------------------------------------
# Helpers
# -----------------------------------------------------------------------------


def _wrap_with_after(
    responder: Responder, action: AfterFn, action_args: Any, action_kwargs: Any
) -> Responder:
    """Execute the given action function after a responder method.

    Args:
        responder: The responder method to wrap.
        action: A function with a signature similar to a resource responder
            method, taking the form ``func(req, resp, resource)``.
        action_args: Additional positional arguments to pass to *action*.
        action_kwargs: Additional keyword arguments to pass to *action*.
    """

    responder_argnames = get_argnames(responder)
    extra_argnames = responder_argnames[2:]  # Skip req, resp
    do_after_responder: Responder

    if iscoroutinefunction(responder):
        async_action = cast('AsyncAfterFn', _wrap_non_coroutine_unsafe(action))
        async_responder = cast('AsgiResponderMethod', responder)

        @wraps(async_responder)
        async def do_after(
            self: Resource,
            req: asgi.Request,
            resp: asgi.Response,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            if args:
                _merge_responder_args(args, kwargs, extra_argnames)

            await async_responder(self, req, resp, **kwargs)
            await async_action(req, resp, self, *action_args, **action_kwargs)

        do_after_responder = cast('AsgiResponderMethod', do_after)
    else:
        sync_action = cast('SyncAfterFn', action)
        sync_responder = cast('ResponderMethod', responder)

        @wraps(sync_responder)
        def do_after(
            self: Resource,
            req: wsgi.Request,
            resp: wsgi.Response,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            if args:
                _merge_responder_args(args, kwargs, extra_argnames)

            sync_responder(self, req, resp, **kwargs)
            sync_action(req, resp, self, *action_args, **action_kwargs)

        do_after_responder = cast('ResponderMethod', do_after)
    return do_after_responder


def _wrap_with_before(
    responder: Responder,
    action: BeforeFn,
    action_args: Tuple[Any, ...],
    action_kwargs: Dict[str, Any],
) -> Responder:
    """Execute the given action function before a responder method.

    Args:
        responder: The responder method to wrap.
        action: A function with a similar signature to a resource responder
            method, taking the form ``func(req, resp, resource, params)``.
        action_args: Additional positional arguments to pass to *action*.
        action_kwargs: Additional keyword arguments to pass to *action*.
    """

    responder_argnames = get_argnames(responder)
    extra_argnames = responder_argnames[2:]  # Skip req, resp
    do_before_responder: Responder

    if iscoroutinefunction(responder):
        async_action = cast('AsyncBeforeFn', _wrap_non_coroutine_unsafe(action))
        async_responder = cast('AsgiResponderMethod', responder)

        @wraps(async_responder)
        async def do_before(
            self: Resource,
            req: asgi.Request,
            resp: asgi.Response,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            if args:
                _merge_responder_args(args, kwargs, extra_argnames)

            await async_action(req, resp, self, kwargs, *action_args, **action_kwargs)
            await async_responder(self, req, resp, **kwargs)

        do_before_responder = cast('AsgiResponderMethod', do_before)
    else:
        sync_action = cast('SyncBeforeFn', action)
        sync_responder = cast('ResponderMethod', responder)

        @wraps(sync_responder)
        def do_before(
            self: Resource,
            req: wsgi.Request,
            resp: wsgi.Response,
            *args: Any,
            **kwargs: Any,
        ) -> None:
            if args:
                _merge_responder_args(args, kwargs, extra_argnames)

            sync_action(req, resp, self, kwargs, *action_args, **action_kwargs)
            sync_responder(self, req, resp, **kwargs)

        do_before_responder = cast('ResponderMethod', do_before)
    return do_before_responder


def _merge_responder_args(
    args: Tuple[Any, ...], kwargs: Dict[str, Any], argnames: List[str]
) -> None:
    """Merge responder args into kwargs.

    The framework always passes extra args as keyword arguments.
    However, when the app calls the responder directly, it might use
    positional arguments instead, so we need to handle that case. This
    might happen, for example, when overriding a resource and calling
    a responder via super().

    Args:
        args (tuple): Extra args passed into the responder
        kwargs (dict): Keyword args passed into the responder
        argnames (list): Extra argnames from the responder's
            signature, ordered as defined
    """

    # NOTE(kgriffs): Merge positional args into kwargs by matching
    # them up to the responder's signature. To do that, we must
    # find out the names of the positional arguments by matching
    # them in the order of the arguments named in the responder's
    # signature.
    for i, argname in enumerate(argnames):
        # NOTE(kgriffs): extra_argnames may contain keyword arguments,
        # which won't be in the args list, and are already in the kwargs
        # dict anyway, so detect and skip them.
        if argname not in kwargs:
            kwargs[argname] = args[i]



# Copyright 2025 by Vytautas Liuolia.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Hacks and workarounds for Sphinx bugs."""

import sphinx.util.inspect


def patch_type_alias_forward_ref(cls: type) -> None:
    def __hash__(obj) -> int:
        return hash(obj.name)

    # NOTE(vytas): Based on various issue comments in the Sphinx repo.
    def __repr__(obj) -> str:
        return obj.name

    cls.__hash__ = __hash__
    cls.__repr__ = __repr__


def setup(app):
    patch_type_alias_forward_ref(sphinx.util.inspect.TypeAliasForwardRef)

    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import pytest

try:
    import httpx
except ImportError:
    httpx = None  # type: ignore

try:
    import requests
except ImportError:
    requests = None  # type: ignore

import falcon.testing as testing


def test_quote(util):
    quote = util.load_module('examples/quote.py')

    resp = testing.simulate_get(quote.app, '/quote')

    assert resp.status_code == 200
    assert resp.json == {
        'author': 'Grace Hopper',
        'quote': "I've always been more interested in the future than in the past.",
    }


def test_things(asgi, util):
    suffix = '_asgi' if asgi else ''
    things = util.load_module(f'examples/things{suffix}.py')

    resp = testing.simulate_get(things.app, '/things')

    assert resp.status_code == 200
    assert resp.text == (
        '\nTwo things awe me most, the starry sky above me and the moral law within me.'
        '\n\n    ~ Immanuel Kant\n\n'
    )


@pytest.mark.skipif(
    httpx is None, reason='things_advanced_asgi.py requires httpx [not found]'
)
@pytest.mark.skipif(
    requests is None, reason='things_advanced.py requires requests [not found]'
)
def test_things_advanced(asgi, util):
    suffix = '_asgi' if asgi else ''
    advanced = util.load_module(f'examples/things_advanced{suffix}.py')

    # NOTE(vytas): The ASGI example explicitly requires Content-Length
    #   (its middleware errors out otherwise with 400).
    #   Should we change this?
    resp1 = testing.simulate_get(
        advanced.app, '/1337/things', headers={'Content-Length': '0'}
    )
    assert resp1.status_code == 401

    resp2 = testing.simulate_get(
        advanced.app,
        '/1337/things',
        headers={'Authorization': 'custom-token', 'Content-Length': '0'},
    )
    assert resp2.status_code == 200
    assert len(resp2.json) == 1
    assert resp2.json[0]['color'] == 'green'



def test_missing_in_store(client):
    resp = client.simulate_get(
        '/thumbnails/1a256a8a-2063-46ff-b53f-d04d5bcf5eee/80x80.jpeg'
    )
    assert resp.status_code == 404


def test_thumbnails(client, png_image, image_size):
    resp1 = client.simulate_post('/images', body=png_image)
    assert resp1.status_code == 201

    thumbnails = resp1.json['thumbnails']
    assert set(thumbnails) == {
        '/thumbnails/36562622-48e5-4a61-be67-e426b11821ed/320x180.jpeg',
        '/thumbnails/36562622-48e5-4a61-be67-e426b11821ed/160x90.jpeg',
    }

    for uri in thumbnails:
        resp = client.simulate_get(uri)
        assert resp.headers['Content-Type'] == 'image/jpeg'
        assert resp.headers['X-ASGILook-Cache'] == 'Miss'
        assert image_size(resp.content) in ((320, 180), (160, 90))


def test_missing_size(client, png_image):
    client.simulate_post('/images', body=png_image)

    resp = client.simulate_get(
        '/thumbnails/36562622-48e5-4a61-be67-e426b11821ed/480x270.jpeg'
    )
    assert resp.status_code == 404


def test_thumbnail_caching(client, png_image):
    client.simulate_post('/images', body=png_image)

    reference = None
    for retry in range(4):
        resp = client.simulate_get(
            '/thumbnails/36562622-48e5-4a61-be67-e426b11821ed/160x90.jpeg'
        )
        assert resp.status_code == 200
        if retry == 0:
            assert resp.headers.get('X-ASGILook-Cache') == 'Miss'
            reference = resp.content
        else:
            assert resp.headers.get('X-ASGILook-Cache') == 'Hit'
            assert resp.content == reference



import aiofiles

import falcon


class Images:
    def __init__(self, config, store):
        self._config = config
        self._store = store

    async def on_get(self, req, resp):
        resp.media = [image.serialize() for image in self._store.list_images()]

    async def on_get_image(self, req, resp, image_id):
        # NOTE: image_id: UUID is converted back to a string identifier.
        image = self._store.get(str(image_id))
        if not image:
            raise falcon.HTTPNotFound

        resp.stream = await aiofiles.open(image.path, 'rb')
        resp.content_type = falcon.MEDIA_JPEG

    async def on_post(self, req, resp):
        data = await req.stream.read()
        image_id = str(self._config.uuid_generator())
        image = await self._store.save(image_id, data)

        resp.location = image.uri
        resp.media = image.serialize()
        resp.status = falcon.HTTP_201


class Thumbnails:
    def __init__(self, store):
        self._store = store

    async def on_get(self, req, resp, image_id, width, height):
        image = self._store.get(str(image_id))
        if not image:
            raise falcon.HTTPNotFound
        if req.path not in image.thumbnails():
            raise falcon.HTTPNotFound

        resp.content_type = falcon.MEDIA_JPEG
        resp.data = await self._store.make_thumbnail(image, (width, height))



# Copyright 2013 by Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functional testing framework for Falcon apps and Falcon itself.

Falcon's testing module contains various test classes and utility
functions to support functional testing for both Falcon-based apps and
the Falcon framework itself.

The testing framework supports both unittest and pytest::

    # -----------------------------------------------------------------
    # unittest
    # -----------------------------------------------------------------

    from falcon import testing
    import myapp


    class MyTestCase(testing.TestCase):
        def setUp(self):
            super(MyTestCase, self).setUp()

            # Assume the hypothetical `myapp` package has a
            # function called `create()` to initialize and
            # return a `falcon.App` instance.
            self.app = myapp.create()


    class TestMyApp(MyTestCase):
        def test_get_message(self):
            doc = {'message': 'Hello world!'}

            result = self.simulate_get('/messages/42')
            self.assertEqual(result.json, doc)


    # -----------------------------------------------------------------
    # pytest
    # -----------------------------------------------------------------

    from falcon import testing
    import pytest

    import myapp


    # Depending on your testing strategy and how your application
    # manages state, you may be able to broaden the fixture scope
    # beyond the default 'function' scope used in this example.

    @pytest.fixture()
    def client():
        # Assume the hypothetical `myapp` package has a function called
        # `create()` to initialize and return a `falcon.App` instance.
        return testing.TestClient(myapp.create())


    def test_get_message(client):
        doc = {'message': 'Hello world!'}

        result = client.simulate_get('/messages/42')
        assert result.json == doc
"""

# Hoist classes and functions into the falcon.testing namespace
from falcon import util as _util
from falcon.testing.client import ASGIConductor
from falcon.testing.client import Cookie
from falcon.testing.client import Result
from falcon.testing.client import ResultBodyStream
from falcon.testing.client import simulate_delete
from falcon.testing.client import simulate_get
from falcon.testing.client import simulate_head
from falcon.testing.client import simulate_options
from falcon.testing.client import simulate_patch
from falcon.testing.client import simulate_post
from falcon.testing.client import simulate_put
from falcon.testing.client import simulate_request
from falcon.testing.client import StreamedResult
from falcon.testing.client import TestClient
from falcon.testing.helpers import ASGILifespanEventEmitter
from falcon.testing.helpers import ASGIRequestEventEmitter
from falcon.testing.helpers import ASGIResponseEventCollector
from falcon.testing.helpers import ASGIWebSocketSimulator
from falcon.testing.helpers import closed_wsgi_iterable
from falcon.testing.helpers import create_asgi_req
from falcon.testing.helpers import create_environ
from falcon.testing.helpers import create_req
from falcon.testing.helpers import create_scope
from falcon.testing.helpers import create_scope_ws
from falcon.testing.helpers import DEFAULT_HOST
from falcon.testing.helpers import DEFAULT_UA
from falcon.testing.helpers import get_encoding_from_headers
from falcon.testing.helpers import get_unused_port
from falcon.testing.helpers import rand_string
from falcon.testing.helpers import redirected
from falcon.testing.resource import capture_responder_args
from falcon.testing.resource import capture_responder_args_async
from falcon.testing.resource import set_resp_defaults
from falcon.testing.resource import set_resp_defaults_async
from falcon.testing.resource import SimpleTestResource
from falcon.testing.resource import SimpleTestResourceAsync
from falcon.testing.srmock import StartResponseMock
from falcon.testing.test_case import TestCase

__all__ = (
    # client
    'ASGIConductor',
    'Cookie',
    'Result',
    'ResultBodyStream',
    'simulate_delete',
    'simulate_get',
    'simulate_head',
    'simulate_options',
    'simulate_patch',
    'simulate_post',
    'simulate_put',
    'simulate_request',
    'StreamedResult',
    'TestClient',
    # helpers
    'ASGILifespanEventEmitter',
    'ASGIRequestEventEmitter',
    'ASGIResponseEventCollector',
    'ASGIWebSocketSimulator',
    'closed_wsgi_iterable',
    'create_asgi_req',
    'create_environ',
    'create_req',
    'create_scope',
    'create_scofrom_headers',
    'get_unused_port',
    'rand_string',
    'redirected',
    # resource
    'capture_responder_args',
    'capture_responder_args_async',
    'set_resp_defaults',
    'set_resp_defaults_async',
    'SimpleTestResource',
    'SimpleTestResourceAsync',
    # srmock
    'StartResponseMock',
    # test_case
    'TestCase',
)


# NOTE(kgriffs): Alias for backwards-compatibility with Falcon 0.2
# TODO(vytas): Remove in Falcon 5.0.
httpnow = _util.deprecated(
    'This method is deprecated and will be removed in Falcon 5.0. '
    'Use `falcon.util.http_now` instead.'
)(_util.http_now)



import asyncio
from collections import Counter
import hashlib
import platform
import sys
import time

import falcon
import falcon.asgi
import falcon.errors
import falcon.util

SSE_TEST_MAX_DELAY_SEC = 1
_WIN32 = sys.platform.startswith('win')
_X86_64 = platform.machine() == 'x86_64'


class Things:
    def __init__(self):
        self._counter = Counter()

    async def on_get(self, req, resp):
        await asyncio.sleep(0.01)
        resp.text = req.remote_addr

    async def on_post(self, req, resp):
        resp.data = await req.stream.read(req.content_length or 0)
        resp.set_header('X-Counter', str(self._counter['background:things:on_post']))

        async def background_job_async():
            await asyncio.sleep(0.01)
            self._counter['background:things:on_post'] += 1

        def background_job_sync():
            time.sleep(0.01)
            self._counter['background:things:on_post'] += 1000

        resp.schedule(background_job_async)
        resp.schedule_sync(background_job_sync)
        resp.schedule(background_job_async)
        resp.schedule_sync(background_job_sync)

    async def on_put(self, req, resp):
        # NOTE(kgriffs): Test that reading past the end does
        # not hang.

        chunks = []
        for i in range(req.content_length + 1):
            # NOTE(kgriffs): In the ASGI interface, bounded_stream is an
            #   alias for req.stream. We'll use the alias here just as
            #   a sanity check.
            chunk = await req.bounded_stream.read(1)
            chunks.append(chunk)

        # NOTE(kgriffs): body should really be set to a string, but
        #   Falcon is lenient and will allow bytes as well (although
        #   it is slightly less performant).
        # TODO(kgriffs): Perhaps in Falcon 4.0 be more strict? We would
        #   also have to change the WSGI behavior to match.
        resp.text = b''.join(chunks)

        # =================================================================
        # NOTE(kgriffs): Test the sync_to_async helpers here to make sure
        #   they work as expected in the context of a real ASGI server.
        # =================================================================
        safely_tasks = []
        safely_values = []

        def callmesafely(a, b, c=None):
            # NOTE(kgriffs): Sleep to prove that there isn't another instance
            #   running in parallel that is able to race ahead.
            time.sleep(0.001)
            safely_values.append((a, b, c))

        cms = falcon.util.wrap_sync_to_async(callmesafely, threadsafe=False)

        # NOTE(caselit): on windows it takes more time so create less tasks
        # NOTE(vytas): Tests on non-x86 platforms are run using software
        #   emulation via single-thread QEMU Docker containers, making them
        #   considerably slower as well.
        num_cms_tasks = 100 if _WIN32 or not _X86_64 else 1000

        for i in range(num_cms_tasks):
            # NOTE(kgriffs): create_task() is used here, so that the coroutines
            #   are scheduled immediately in the order created; under Python
            #   3.6, asyncio.gather() does not seem to always schedule
            #   them in order, so we do it this way to make it predictable.
            safely_tasks.append(asyncio.create_task(cms(i, i + 1, c=i + 2)))

        await asyncio.gather(*safely_tasks)

        assert len(safely_values) == num_cms_tasks
        for i, val in enumerate(safely_values):
            assert safely_values[i] == (i, i + 1, i + 2)

        def callmeshirley(a=42, b=None):
            return (a, b)

        assert (42, None) == await falcon.util.sync_to_async(callmeshirley)
        assert (1, 2) == await falcon.util.sync_to_async(callmeshirley, 1, 2)
        assert (5, None) == await falcon.util.sync_to_async(callmeshirley, 5)
        assert (3, 4) == await falcon.util.sync_to_async(callmeshirley, 3, b=4)


class Bucket:
    async def on_post(self, req, resp):
        resp.text = await req.stream.read()

    async def on_put_drops(self, req, resp):
        # NOTE(kgriffs): Integrity check
        sha1 = hashlib.sha1()

        drops = 0
        async for drop in req.stream:
            drops += 1
            sha1.update(drop)

        resp.media = {'drops': drops, 'sha1': sha1.hexdigest()}


class Feed:
    async def on_websocket(self, req, ws, feed_id):
        await ws.accept()
        await ws.send_text(feed_id)


class Events:
    async def on_get(self, req, resp):
        async def emit():
            s = 0
            while s <= SSE_TEST_MAX_DELAY_SEC:
                yield falcon.asgi.SSEvent(text='hello world')
                await asyncio.sleep(s)
                s += SSE_TEST_MAX_DELAY_SEC / 4

        resp.sse = emit()

    async def on_websocket(self, req, ws):  # noqa: C901
        recv_command = req.get_header('X-Command') == 'recv'
        send_mismatched = req.get_header('X-Mismatch') == 'send'
        recv_mismatched = req.get_header('X-Mismatch') == 'recv'

        mismatch_type = req.get_header('X-Mismatch-Type', default='text')

        raise_error = req.get_header('X-Raise-Error')

        close = req.get_header('X-Close')
        close_code = req.get_header('X-Close-Code')
        if close_code:
            close_code = int(close_code)

        accept = req.get_header('X-Accept', default='accept')

        if accept == 'accept':
            subprotocol = req.get_header('X-Subprotocol')

            if subprotocol == '*':
                subprotocol = ws.subprotocols[0]

            if subprotocol:
                await ws.accept(subprotocol)
            else:
                await ws.accept()
        elif accept == 'reject':
            if close:
                await ws.close()
            return

        if send_mismatched:
            if mismatch_type == 'text':
                await ws.send_text(b'fizzbuzz')
            else:
                await ws.send_data('fizzbuzz')

        if recv_mismatched:
            if mismatch_type == 'text':
                await ws.receive_text()
            else:
                await ws.receive_data()

        start = time.time()
        while time.time() - start < 1:
            try:
                msg = None

                if recv_command:
                    msg = await ws.receive_media()
                else:
                    msg = None

                await ws.send_text('hello world')
                print('on_websocket:send_text')

                if msg and msg['command'] == 'echo':
                    await ws.send_text(msg['echo'])

                await ws.send_data(b'hello\x00world')
                await asyncio.sleep(0.2)
            except falcon.errors.WebSocketDisconnected:
                print('on_websocket:WebSocketDisconnected')
                raise

            if raise_error == 'generic':
                raise Exception('Test: Generic Unhandled Error')
            elif raise_error == 'http':
                raise falcon.HTTPBadRequest()

        if close:
            # NOTE(kgriffs): Tests that the default is used
            #   when close_code is None.
            await ws.close(close_code)


class Multipart:
    async def on_post(self, req, resp):
        parts = {}

        form = await req.get_media()
        async for part in form:
            # NOTE(vytas): SHA1 is no longer recommended for cryptographic
            #   purposes, but here we are only using it for integrity checking.
            sha1 = hashlib.sha1()
            async for chunk in part.stream:
                sha1.update(chunk)

            parts[part.name] = {
                'filename': part.filename,
                'sha1': sha1.hexdigest(),
            }

        resp.media = parts


class LifespanHandler:
    def __init__(self):
        self.startup_succeeded = False
        self.shutdown_succeeded = False

    async def process_startup(self, scope, event):
        assert scope['type'] == 'lifespan'
        assert event['type'] == 'lifespan.startup'
        self.startup_succeeded = True

    async def process_shutdown(self, scope, event):
        assert scope['type'] == 'lifespan'
        assert event['type'] == 'lifespan.shutdown'
        self.shutdown_succeeded = True


class TestJar:
    async def on_get(self, req, resp):
        # NOTE(myusko): In the future we shouldn't change the cookie
        #             a test depends on the input.
        # NOTE(kgriffs): This is the only test that uses a single
        #   cookie (vs. multiple) as input; if this input ever changes,
        #   a separate test will need to be added to explicitly verify
        #   this use case.
        resp.set_cookie('has_permission', 'true')

    async def on_post(self, req, resp):
        if req.cookies['has_permission'] == 'true':
            resp.status = falcon.HTTP_200
        else:
            resp.status = falcon.HTTP_403


class WSOptions:
    _SUPPORTED_KEYS = frozenset(
        {'default_close_reasons', 'error_close_code', 'max_receive_queue'}
    )

    def __init__(self, ws_options):
        self._ws_options = ws_options

    async def on_get(self, req, resp):
        resp.media = {
            key: getattr(self._ws_options, key) for key in self._SUPPORTED_KEYS
        }

    async def on_patch(self, req, resp):
        update = await req.get_media()
        for key, value in update.items():
            if key not in self._SUPPORTED_KEYS:
                raise falcon.HTTPInvalidParam('unsupported option', key)
            setattr(self._ws_options, key, value)

        resp.status = falcon.HTTP_NO_CONTENT


def create_app():
    app = falcon.asgi.App()
    bucket = Bucket()
    lifespan_handler = LifespanHandler()

    app.add_route('/', Things())
    app.add_route('/bucket', bucket)
    app.add_route('/bucket/drops', bucket, suffix='drops')
    app.add_route('/events', Events())
    app.add_route('/forms', Multipart())
    app.add_route('/jars', TestJar())
    app.add_route('/feeds/{feed_id}', Feed())
    app.add_route('/wsoptions', WSOptions(app.ws_options))

    app.add_middleware(lifespan_handler)

    async def _on_ws_error(req, resp, error, params, ws=None):
        if not ws:
            raise

        if ws.unaccepted:
            await ws.accept()

        if not ws.closed:
            await ws.send_text(error.__class__.__name__)
            await ws.close()

    app.add_error_handler(falcon.errors.OperationNotAllowed, _on_ws_error)
    app.add_error_handler(ValueError, _on_ws_error)

    return app


application = create_app()



from datetime import datetime
from datetime import timezone
import math
import string
import uuid

import pytest

from falcon.routing import converters

_TEST_UUID = uuid.uuid4()
_TEST_UUID_STR = str(_TEST_UUID)
_TEST_UUID_STR_SANS_HYPHENS = _TEST_UUID_STR.replace('-', '')


def _as_params(*values, prefix=''):
    # NOTE(caselit): each value must be a tuple/list even when using one
    #   single argument
    return [
        pytest.param(*value, id=f'{prefix}_{i}' if prefix else f'{i}')
        for i, value in enumerate(values, 1)
    ]


@pytest.mark.parametrize(
    'value, num_digits, min, max, expected',
    [
        ('123', None, None, None, 123),
        ('01', None, None, None, 1),
        ('001', None, None, None, 1),
        ('0', None, None, None, 0),
        ('00', None, None, None, 00),
        ('1', 1, None, None, 1),
        ('12', 1, None, None, None),
        ('12', 2, None, None, 12),
        ('1', 1, 1, 1, 1),
        ('1', 1, 1, None, 1),
        ('1', 1, 1, 2, 1),
        ('1', 1, 2, None, None),
        ('1', 1, 2, 1, None),
        ('2', 1, 1, 2, 2),
        ('2', 1, 2, 2, 2),
        ('3', 1, 1, 2, None),
        ('12', 1, None, None, None),
        ('12', 1, 1, 12, None),
        ('12', 2, None, None, 12),
        ('12', 2, 1, 12, 12),
        ('12', 2, 12, 12, 12),
        ('12', 2, 13, 12, None),
        ('12', 2, 13, 13, None),
    ],
)
def test_int_converter(value, num_digits, min, max, expected):
    c = converters.IntConverter(num_digits, min, max)
    assert c.convert(value) == expected


@pytest.mark.parametrize(
    'value',
    (
        ['0x0F', 'something', '', ' ']
        + ['123' + w for w in string.whitespace]
        + [w + '123' for w in string.whitespace]
    ),
)
def test_int_converter_malformed(value):
    c = converters.IntConverter()
    assert c.convert(value) is None


@pytest.mark.parametrize('num_digits', [0, -1, -10])
def test_int_converter_invalid_config(num_digits):
    with pytest.raises(ValueError):
        converters.IntConverter(num_digits)


@pytest.mark.parametrize(
    'value, min, max, expected',
    [
        ('123', None, None, 123.0),
        ('01', None, None, 1.0),
        ('001', None, None, 1.0),
        ('0', None, None, 0.0),
        ('00', None, None, 00.0),
        ('1', None, None, 1.0),
        ('12', 2, None, 12.0),
        ('1', 1, 1, 1.0),
        ('12', 1, 20, 12.0),
        ('12', 2, 10, None),
        ('1', 2, 13, None),
        ('-1', 2, 13, None),
        ('-1', -2, 10, -1.0),
        ('1.4', 1, 10, 1.4),
        ('inf', 1, 100, None),
        ('-inf', 1, 1000, None),
        ('nan', 1, 1000, None),
        ('1.5e100', 0, 1, None),
        ('0.5e1', 0, 10, 5.0),
        ('-1.6e1', -50, 50, -16.0),
        ('1.5e2', 0, 500, 150),
        ('-1.6e10', -1.7e10, 1.0e10, -16000000000.0),
    ],
)
def test_float_converter(value, min, max, expected):
    c = converters.FloatConverter(min, max)
    assert c.convert(value) == expected


@pytest.mark.parametrize(
    'value', ['nan', 'NaN', 'NAN', 'nAn', '-inf', 'inf', '-INF', 'INF']
)
def test_float_converter_non_finite_allowed(value):
    c = converters.FloatConverter(finite=False)
    assert not math.isfinite(c.convert(value))


@pytest.mark.parametrize('finite', [None, True])
def test_float_converter_non_finite_disallowed(finite):
    c = converters.FloatConverter(finite=finite)
    assert c.convert('nan') is None


@pytest.mark.parametrize(
    'value',
    (
        ['0x0F', 'something', '', ' ']
        + ['123' + w for w in string.whitespace]
        + [w + '123' for w in string.whitespace]
    ),
)
def test_float_converter_malformed(value):
    c = converters.FloatConverter()
    assert c.convert(value) is None


@pytest.mark.parametrize(
    'value, format_string, expected',
    [
        ('07-03-17', '%m-%d-%y', datetime(2017, 7, 3)),
        ('07-03-17 ', '%m-%d-%y ', datetime(2017, 7, 3)),
        ('2017-07-03T14:30:01Z', '%Y-%m-%dT%H:%M:%SZ', datetime(2017, 7, 3, 14, 30, 1)),
        ('2017-07-03T14:30:01', '%Y-%m-%dT%H:%M:%S', datetime(2017, 7, 3, 14, 30, 1)),
        ('2017_19', '%Y_%H', datetime(2017, 1, 1, 19, 0)),
        ('2017-07-03T14:30:01', '%Y-%m-%dT%H:%M:%SZ', None),
        ('07-03-17 ', '%m-%d-%y', None),
        (' 07-03-17', '%m-%d-%y', None),
        ('07 -03-17', '%m-%d-%y', None),
    ],
)
def test_datetime_converter(value, format_string, expected):
    c = converters.DateTimeConverter(format_string)
    assert c.convert(value) == expected


def test_datetime_converter_default_format():
    c = converters.DateTimeConverter()
    assert c.convert('2017-07-03T14:30:01Z') == datetime(
        2017, 7, 3, 14, 30, 1, tzinfo=timezone.utc
    )


@pytest.mark.parametrize(
    'value, expected',
    _as_params(
        (_TEST_UUID_STR, _TEST_UUID),
        (_TEST_UUID_STR.replace('-', '', 1), _TEST_UUID),
        (_TEST_UUID_STR_SANS_HYPHENS, _TEST_UUID),
        ('urn:uuid:' + _TEST_UUID_STR, _TEST_UUID),
        ('urn:uuid:' + _TEST_UUID_STR_SANS_HYPHENS, _TEST_UUID),
        (' ', None),
        (_TEST_UUID_STR + ' ', None),
        (' ' + _TEST_UUID_STR, None),
        (_TEST_UUID_STR[:-1], None),
        (_TEST_UUID_STR[0], None),
        (_TEST_UUID_STR[:-1] + 'g', None),
        (_TEST_UUID_STR.replace('-', '_'), None),
        prefix='uuid',
    ),
)
def test_uuid_converter(value, expected):
    c = converters.UUIDConverter()
    assert c.convert(value) == expected
